from basetest import *


class TestExternalFailures(unittest2.TestCase):

    def setUp(self):
        prepare_zpools()
        self.longMessage = True

    # generate a resumable state
    # NOTE: this generates two resumable test_target1/test_source1/fs1 and test_target1/test_source1/fs1/sub
    def generate_resume(self):

        r = shelltest("zfs set compress=off test_source1 test_target1")

        # big change on source
        r = shelltest("dd if=/dev/zero of=/test_source1/fs1/data bs=250M count=1")

        # waste space on target
        r = shelltest("dd if=/dev/zero of=/test_target1/waste bs=250M count=1")

        # should fail and leave resume token (if supported)
        self.assertTrue(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())

        # free up space
        r = shelltest("rm /test_target1/waste")
        # sync
        r = shelltest("zfs umount test_target1")
        r = shelltest("zfs mount test_target1")

    # resume initial backup
    def test_initial_resume(self):

        # inital backup, leaves resume token
        with mocktime("20101111000000"):
            self.generate_resume()

        # --test should resume and succeed
        with OutputIO() as buf:
            with redirect_stdout(buf):
                self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --test".split(" ")).run())

            print(buf.getvalue())

            self.assertIn(": resuming", buf.getvalue())

        # should resume and succeed
        with OutputIO() as buf:
            with redirect_stdout(buf):
                self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())

            print(buf.getvalue())

            self.assertIn(": resuming", buf.getvalue())

        r = shelltest("zfs list -H -o name -r -t all test_target1")
        self.assertMultiLineEqual(r, """
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
""")

    # resume incremental backup
    def test_incremental_resume(self):

        # initial backup
        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty".split(" ")).run())

        # incremental backup leaves resume token
        with mocktime("20101111000001"):
            self.generate_resume()

        # --test should resume and succeed
        with OutputIO() as buf:
            with redirect_stdout(buf):
                self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --test".split(" ")).run())

            print(buf.getvalue())

            self.assertIn(": resuming", buf.getvalue())

        # should resume and succeed
        with OutputIO() as buf:
            with redirect_stdout(buf):
                self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())

            print(buf.getvalue())

            # did we really resume?
            self.assertIn(": resuming", buf.getvalue())

        r = shelltest("zfs list -H -o name -r -t all test_target1")
        self.assertMultiLineEqual(r, """
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
""")

    # generate an invalid resume token, and verify if its aborted automaticly
    def test_initial_resumeabort(self):


        # inital backup, leaves resume token
        with mocktime("20101111000000"):
            self.generate_resume()

        # remove corresponding source snapshot, so it becomes invalid
        shelltest("zfs destroy test_source1/fs1@test-20101111000000")

        # NOTE: it can only abort the initial dataset if it has no subs
        shelltest("zfs destroy test_target1/test_source1/fs1/sub; true")

        # --test try again, should abort old resume
        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --test".split(" ")).run())

        # try again, should abort old resume
        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())

        r = shelltest("zfs list -H -o name -r -t all test_target1")
        self.assertMultiLineEqual(r, """
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
""")

    # generate an invalid resume token, and verify if its aborted automaticly
    def test_incremental_resumeabort(self):

        # initial backup
        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty".split(" ")).run())

        # icremental backup, leaves resume token
        with mocktime("20101111000001"):
            self.generate_resume()

        # remove corresponding source snapshot, so it becomes invalid
        shelltest("zfs destroy test_source1/fs1@test-20101111000001")

        # --test try again, should abort old resume
        with mocktime("20101111000002"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --test".split(" ")).run())

        # try again, should abort old resume
        with mocktime("20101111000002"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())

        r = shelltest("zfs list -H -o name -r -t all test_target1")
        self.assertMultiLineEqual(r, """
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@test-20101111000002
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
""")

    # create a resume situation, where the other side doesnt want the snapshot anymore ( should abort resume )
    def test_abort_unwanted_resume(self):

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())

        # generate resume
        with mocktime("20101111000001"):
            self.generate_resume()

        with OutputIO() as buf:
            with redirect_stdout(buf):
                # incremental, doesnt want previous anymore
                with mocktime("20101111000002"):
                    self.assertFalse(ZfsAutobackup(
                        "test test_target1 --no-progress --verbose --keep-target=0 --allow-empty --debug".split(" ")).run())

            print(buf.getvalue())

            self.assertIn("Aborting resume, we dont want that snapshot anymore.", buf.getvalue())

        r = shelltest("zfs list -H -o name -r -t all test_target1")
        self.assertMultiLineEqual(r, """
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000002
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000002
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000002
""")

    # test with empty snapshot list (this was a bug)
    def test_abort_resume_emptysnapshotlist(self):

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())

        # generate resume
        with mocktime("20101111000001"):
            self.generate_resume()

        shelltest("zfs destroy test_source1/fs1@test-20101111000001")

        with OutputIO() as buf:
            with redirect_stdout(buf):
                # incremental, doesnt want previous anymore
                with mocktime("20101111000002"):
                    self.assertFalse(ZfsAutobackup(
                        "test test_target1 --no-progress --verbose --no-snapshot".split(
                            " ")).run())

            print(buf.getvalue())

            self.assertIn("Aborting resume, its obsolete", buf.getvalue())


    def test_missing_common(self):

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty".split(" ")).run())

        # remove common snapshot and leave nothing
        shelltest("zfs release zfs_autobackup:test test_source1/fs1@test-20101111000000")
        shelltest("zfs destroy test_source1/fs1@test-20101111000000")

        with mocktime("20101111000001"):
            self.assertTrue(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty".split(" ")).run())

    #UPDATE: offcourse the one thing that wasn't tested had a bug :(  (in ExecuteNode.run()).
    def test_ignoretransfererrors(self):

            self.skipTest("Not sure how to implement a test for this without some serious hacking and patching.")

#         #recreate target pool without any features
#         # shelltest("zfs set compress=on test_source1; zpool destroy test_target1; zpool create test_target1 -o feature@project_quota=disabled /dev/ram2")
#
#         with mocktime("20101111000000"):
#             self.assertFalse(ZfsAutobackup("test test_target1 --verbose --allow-empty --no-progress".split(" ")).run())
#
#         r = shelltest("zfs list -H -o name -r -t all test_target1")
#
#         self.assertMultiLineEqual(r, """
# test_target1
# test_target1/test_source1
# test_target1/test_source1/fs1
# test_target1/test_source1/fs1@test-20101111000002
# test_target1/test_source1/fs1/sub
# test_target1/test_source1/fs1/sub@test-20101111000002
# test_target1/test_source2
# test_target1/test_source2/fs2
# test_target1/test_source2/fs2/sub
# test_target1/test_source2/fs2/sub@test-20101111000002
#         """)



from basetest import *
from zfs_autobackup.ExecuteNode import *

print("THIS TEST REQUIRES SSH TO LOCALHOST")

class TestExecuteNode(unittest2.TestCase):

    # def setUp(self):

    #     return super().setUp()

    def basics(self, node ):

        with self.subTest("simple echo"):
            self.assertEqual(node.run(["echo","test"]), ["test"])

        with self.subTest("error exit code"):
            with self.assertRaises(ExecuteError):
                node.run(["false"])

        #
        with self.subTest("multiline without tabsplit"):
            self.assertEqual(node.run(["echo","l1c1\tl1c2\nl2c1\tl2c2"], tab_split=False), ["l1c1\tl1c2", "l2c1\tl2c2"])

        #multiline tabsplit
        with self.subTest("multiline tabsplit"):
            self.assertEqual(node.run(["echo","l1c1\tl1c2\nl2c1\tl2c2"], tab_split=True), [['l1c1', 'l1c2'], ['l2c1', 'l2c2']])

        #escaping test
        with self.subTest("escape test"):
            s="><`'\"@&$()$bla\\/.* !#test _+-={}[]|${bla} $bla"
            self.assertEqual(node.run(["echo",s]), [s])

        #return std err as well, trigger stderr by listing something non existing
        with self.subTest("stderr return"):
            (stdout, stderr)=node.run(["sh", "-c", "echo bla >&2"], return_stderr=True, valid_exitcodes=[0])
            self.assertEqual(stdout,[])
            self.assertRegex(stderr[0],"bla")

        #slow command, make sure things dont exit too early
        with self.subTest("early exit test"):
            start_time=time.time()
            self.assertEqual(node.run(["sleep","1"]), [])
            self.assertGreaterEqual(time.time()-start_time,1)

        #input a string and check it via cat
        with self.subTest("stdin input string"):
            self.assertEqual(node.run(["cat"], inp="test"), ["test"])

        #command that wants input, while we dont have input, shouldnt hang forever.
        with self.subTest("stdin process with inp=None (shouldn't hang)"):
            self.assertEqual(node.run(["cat"]), [])

        # let the system do the piping with an unescaped |:
        with self.subTest("system piping test"):

            #first make sure the actual | character is still properly escaped:
            self.assertEqual(node.run(["echo","|"]), ["|"])

            #now pipe
            self.assertEqual(node.run(["echo", "abc", node.PIPE, "tr", "a", "A" ]), ["Abc"])

    def test_basics_local(self):
        node=ExecuteNode(debug_output=True)
        self.basics(node)

    def test_basics_remote(self):
        node=ExecuteNode(ssh_to="localhost", debug_output=True)
        self.basics(node)

    ################

    def test_readonly(self):
        node=ExecuteNode(debug_output=True, readonly=True)

        self.assertEqual(node.run(["echo","test"], readonly=False), [])
        self.assertEqual(node.run(["echo","test"], readonly=True), ["test"])


    ################

    def pipe(self, nodea, nodeb):

        with self.subTest("pipe data"):
            output=nodea.run(["dd", "if=/dev/zero", "count=1000"],pipe=True)
            self.assertEqual(nodeb.run(["md5sum"], inp=output), ["816df6f64deba63b029ca19d880ee10a  -"])

        with self.subTest("exit code both ends of pipe ok"):
            output=nodea.run(["true"], pipe=True)
            nodeb.run(["true"], inp=output)

        with self.subTest("error on pipe input side"):
            with self.assertRaises(ExecuteError):
                output=nodea.run(["false"], pipe=True)
                nodeb.run(["true"], inp=output)

        with self.subTest("error on both sides, ignore exit codes"):
            output=nodea.run(["false"], pipe=True, valid_exitcodes=[])
            nodeb.run(["false"], inp=output, valid_exitcodes=[])

        with self.subTest("error on pipe output side "):
            with self.assertRaises(ExecuteError):
                output=nodea.run(["true"], pipe=True)
                nodeb.run(["false"], inp=output)

        with self.subTest("error on both sides of pipe"):
            with self.assertRaises(ExecuteError):
                output=nodea.run(["false"], pipe=True)
                nodeb.run(["false"], inp=output)

        with self.subTest("check stderr on pipe output side"):
            output=nodea.run(["true"], pipe=True, valid_exitcodes=[0])
            (stdout, stderr)=nodeb.run(["sh", "-c", "echo bla >&2"], inp=output, return_stderr=True, valid_exitcodes=[0])
            self.assertEqual(stdout,[])
            self.assertRegex(stderr[0], "bla" )

        with self.subTest("check stderr on pipe input side (should be only printed)"):
            output=nodea.run(["sh", "-c", "echo bla >&2"], pipe=True, valid_exitcodes=[0])
            (stdout, stderr)=nodeb.run(["true"], inp=output, return_stderr=True, valid_exitcodes=[0])
            self.assertEqual(stdout,[])
            self.assertEqual(stderr,[])


    def test_pipe_local_local(self):
        nodea=ExecuteNode(debug_output=True)
        nodeb=ExecuteNode(debug_output=True)
        self.pipe(nodea, nodeb)

    def test_pipe_remote_remote(self):
        nodea=ExecuteNode(ssh_to="localhost", debug_output=True)
        nodeb=ExecuteNode(ssh_to="localhost", debug_output=True)
        self.pipe(nodea, nodeb)

    def test_pipe_local_remote(self):
        nodea=ExecuteNode(debug_output=True)
        nodeb=ExecuteNode(ssh_to="localhost", debug_output=True)
        self.pipe(nodea, nodeb)

    def test_pipe_remote_local(self):
        nodea=ExecuteNode(ssh_to="localhost", debug_output=True)
        nodeb=ExecuteNode(debug_output=True)
        self.pipe(nodea, nodeb)


    def test_cwd(self):

        nodea=ExecuteNode(ssh_to="localhost", debug_output=True)
        nodeb=ExecuteNode(debug_output=True)

        #change to a directory with a space and execute a system pipe, check if all piped commands are executed in correct directory.
        shelltest("mkdir '/tmp/space test' 2>/dev/null; true")
        self.assertEqual(nodea.run(cmd=["pwd", ExecuteNode.PIPE, "cat"], cwd="/tmp/space test"), ["/tmp/space test"])
        self.assertEqual(nodea.run(cmd=["cat", ExecuteNode.PIPE, "pwd"], cwd="/tmp/space test"), ["/tmp/space test"])
        self.assertEqual(nodeb.run(cmd=["pwd", ExecuteNode.PIPE, "cat"], cwd="/tmp/space test"), ["/tmp/space test"])
        self.assertEqual(nodeb.run(cmd=["cat", ExecuteNode.PIPE, "pwd"], cwd="/tmp/space test"), ["/tmp/space test"])

    def test_script_handlers(self):

        def test(node):
            results = []
            node.script(lines=["echo line1", "echo line2 1>&2", "exit 123"],
                                  stdout_handler=lambda line: results.append(line),
                                  stderr_handler=lambda line: results.append(line),
                                  exit_handler=lambda exit_code: results.append(exit_code),
                                  valid_exitcodes=[123]
                                  )

            self.assertEqual(results, ["line1", "line2", 123 ])

        with self.subTest("remote"):
            test(ExecuteNode(ssh_to="localhost", debug_output=True))
        #
        with self.subTest("local"):
            test(ExecuteNode(debug_output=True))

    def test_script_defaults(self):

        result=[]
        nodea=ExecuteNode(debug_output=True)
        nodea.script(lines=["echo test"], stdout_handler=lambda line: result.append(line))

        self.assertEqual(result, ["test"])

    def test_script_pipe(self):

        result=[]
        nodea=ExecuteNode()
        cmd_pipe=nodea.script(lines=["echo test"], pipe=True)
        nodea.script(lines=["tr e E"], inp=cmd_pipe,stdout_handler=lambda line: result.append(line))

        self.assertEqual(result, ["tEst"])


    def test_mixed(self):

        #should be able to mix run() and script()
        node=ExecuteNode()

        result=[]
        pipe=node.run(["echo", "test"], pipe=True)
        node.script(["tr e E"], inp=pipe, stdout_handler=lambda line: result.append(line))

        self.assertEqual(result, ["tEst"])









import os
import select
import subprocess
from .CmdPipe import CmdPipe, CmdItem
from .LogStub import LogStub

try:
    from shlex import quote as cmd_quote
except ImportError:
    from pipes import quote as cmd_quote


class ExecuteError(Exception):
    pass


class ExecuteNode(LogStub):
    """an endpoint to execute local or remote commands via ssh"""

    PIPE=1

    def __init__(self, ssh_config=None, ssh_to=None, readonly=False, debug_output=False):
        """ssh_config: custom ssh config
           ssh_to: server you want to ssh to. none means local
           readonly: only execute commands that don't make any changes (useful for testing-runs)
           debug_output: show output and exit codes of commands in debugging output.
        """

        self.ssh_config = ssh_config
        self.ssh_to = ssh_to
        self.readonly = readonly
        self.debug_output = debug_output

    def __repr__(self):
        if self.ssh_to is None:
            return "(local)"
        else:
            return self.ssh_to

    def _parse_stdout(self, line):
        """parse stdout. can be overridden in subclass"""
        if self.debug_output:
            self.debug("STDOUT > " + line.rstrip())

    def _parse_stderr(self, line, hide_errors):
        """parse stderr. can be overridden in subclass"""
        if hide_errors:
            self.debug("STDERR > " + line.rstrip())
        else:
            self.error("STDERR > " + line.rstrip())

    def _quote(self, cmd):
        """return quoted version of command. if it has value PIPE it will add an actual | """
        if cmd==self.PIPE:
            return('|')
        else:
            return cmd_quote(cmd)

    def _shell_cmd(self, cmd, cwd):
        """prefix specified ssh shell to command and escape shell characters"""

        ret=[]

        #add remote shell
        if not self.is_local():
            #note: dont escape this part (executed directly without shell)
            ret=["ssh"]

            if self.ssh_config is not None:
                ret.extend(["-F", self.ssh_config])

            ret.append(self.ssh_to)

        #note: DO escape from here, executed in either local or remote shell.

        shell_str=""

        #add cwd change?
        if cwd is not None:
            shell_str=shell_str + "cd " + self._quote(cwd) + "; "

        shell_str=shell_str + " ".join(map(self._quote, cmd))

        ret.append(shell_str)

        return ret

    def is_local(self):
        return self.ssh_to is None

    def run(self, cmd, inp=None, tab_split=False, valid_exitcodes=None, readonly=False, hide_errors=False,
            return_stderr=False, pipe=False, return_all=False, cwd=None):
        """run a command on the node , checks output and parses/handle output and returns it

        Takes care of proper quoting/escaping/ssh and logging of stdout/err/exit codes.

        Either uses a local shell (sh -c) or remote shell (ssh) to execute the command.
        Therefore the command can have stuff like actual pipes in it, if you dont want to use pipe=True to pipe stuff.

        :param cmd: the actual command, should be a list, where the first item is the command
                    and the rest are parameters. use ExecuteNode.PIPE to add an unescaped |
                    (if you want to use system piping instead of python piping)
        :param pipe: return CmdPipe instead of executing it. (pipe this into another run() command via inp=...)
        :param inp: Can be None, a string or a CmdPipe that was previously returned.
        :param tab_split: split tabbed files in output into a list
        :param valid_exitcodes: list of valid exit codes for this command. Use [] to accept all exit codes. Default [0]
        :param readonly: make this True if the command doesn't make any changes and is safe to execute in testmode
        :param hide_errors: don't show stderr output as error, instead show it as debugging output (use to hide expected errors)
        :param return_stderr: return both stdout and stderr as a tuple. (normally only returns stdout)
        :param return_all: return both stdout and stderr and exit_code as a tuple. (normally only returns stdout)
        :param cwd: Change current working directory before executing command.

        """

        # create new pipe?
        if not isinstance(inp, CmdPipe):
            cmd_pipe = CmdPipe(self.readonly, inp)
        else:
            # add stuff to existing pipe
            cmd_pipe = inp

        # stderr parser
        error_lines = []
        returned_exit_code=None

        def stderr_handler(line):
            if tab_split:
                error_lines.append(line.rstrip().split('\t'))
            else:
                error_lines.append(line.rstrip())
            self._parse_stderr(line, hide_errors)

        # exit code hanlder
        if valid_exitcodes is None:
            valid_exitcodes = [0]

        def exit_handler(exit_code):
            if self.debug_output:
                self.debug("EXIT   > {}".format(exit_code))

            if (valid_exitcodes != []) and (exit_code not in valid_exitcodes):
                self.error("Command \"{}\" returned exit code {} (valid codes: {})".format(cmd_item, exit_code, valid_exitcodes))
                return False

            return True

        # stdout parser
        output_lines = []

        if pipe:
            # dont specify output handler, so it will get piped to next process
            stdout_handler=None
        else:
            # handle output manually, dont pipe it
            def stdout_handler(line):
                if tab_split:
                    output_lines.append(line.rstrip().split('\t'))
                else:
                    output_lines.append(line.rstrip())
                self._parse_stdout(line)

        # add shell command and handlers to pipe
        cmd_item=CmdItem(cmd=self._shell_cmd(cmd, cwd), readonly=readonly, stderr_handler=stderr_handler, exit_handler=exit_handler, shell=self.is_local(), stdout_handler=stdout_handler)
        cmd_pipe.add(cmd_item)

        # return CmdPipe instead of executing?
        if pipe:
            return cmd_pipe

        if cmd_pipe.should_execute():
            self.debug("CMD    > {}".format(cmd_pipe))
        else:
            self.debug("CMDSKIP> {}".format(cmd_pipe))

        # execute and calls handlers in CmdPipe
        if not cmd_pipe.execute():
            raise(ExecuteError("Last command returned error"))

        if return_all:
            return output_lines, error_lines, cmd_item.process and cmd_item.process.returncode
        elif return_stderr:
            return output_lines, error_lines
        else:
            return output_lines

    def script(self, lines, inp=None, stdout_handler=None, stderr_handler=None, exit_handler=None, valid_exitcodes=None, readonly=False, hide_errors=False, pipe=False):
        """Run a multiline script on the node.

        This is much more low level than run() and allows for finer grained control.

        Either uses a local shell (sh -c) or remote shell (ssh) to execute the command.
        You need to do your own escaping/quoting.
        It will do logging of stderr and exit codes, but you should
        specify your stdout handler when calling CmdPipe.execute.
        Also specify the optional stderr/exit code handlers if you need them.
        Handlers are called for each line.
        It wont collect lines internally like run() does, so streams of data can be of unlimited size.

        :param lines: list of lines of the actual script.
        :param inp: Can be None, a string or a CmdPipe that was previously returned.
        :param readonly: make this True if the command doesn't make any changes and is safe to execute in testmode
        :param valid_exitcodes: list of valid exit codes for this command. Use [] to accept all exit codes. Default [0]
        :param hide_errors: don't show stderr output as error, instead show it as debugging output (use to hide expected errors)
        :param pipe: return CmdPipe instead of executing it. (pipe this into another run() command via inp=...)

        """

        # create new pipe?
        if not isinstance(inp, CmdPipe):
            cmd_pipe = CmdPipe(self.readonly, inp)
        else:
            # add stuff to existing pipe
            cmd_pipe = inp

        internal_stdout_handler=None
        if stdout_handler is not None:
            if self.debug_output:
                def internal_stdout_handler(line):
                    self.debug("STDOUT > " + line.rstrip())
                    stdout_handler(line)
            else:
                internal_stdout_handler=stdout_handler

        def internal_stderr_handler(line):
            self._parse_stderr(line, hide_errors)
            if stderr_handler is not None:
                stderr_handler(line)

        # exit code hanlder
        if valid_exitcodes is None:
            valid_exitcodes = [0]

        def internal_exit_handler(exit_code):
            if self.debug_output:
                self.debug("EXIT   > {}".format(exit_code))

            if exit_handler is not None:
                exit_handler(exit_code)

            if (valid_exitcodes != []) and (exit_code not in valid_exitcodes):
                self.error("Script returned exit code {} (valid codes: {})".format(exit_code, valid_exitcodes))
                return False

            return True

        #build command
        cmd=[]

        #add remote shell
        if not self.is_local():
            #note: dont escape this part (executed directly without shell)
            cmd.append("ssh")

            if self.ssh_config is not None:
                cmd.append(["-F", self.ssh_config])

            cmd.append(self.ssh_to)

        # convert to script
        cmd.append("\n".join(lines))

        # add shell command and handlers to pipe
        cmd_item=CmdItem(cmd=cmd, readonly=readonly, stderr_handler=internal_stderr_handler, exit_handler=internal_exit_handler, stdout_handler=internal_stdout_handler, shell=self.is_local())
        cmd_pipe.add(cmd_item)

        self.debug("SCRIPT > {}".format(cmd_pipe))

        if pipe:
            return cmd_pipe
        else:
            return cmd_pipe.execute()




from basetest import *


class TestZfsNode(unittest2.TestCase):

    def setUp(self):
        prepare_zpools()
        self.longMessage=True



    def  test_destroymissing(self):

        #initial backup
        with mocktime("19101111000000"): #1000 years in past
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-holds".split(" ")).run())

        with mocktime("20101111000000"): #far in past
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-holds --allow-empty".split(" ")).run())


        with self.subTest("Should do nothing yet"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-snapshot --destroy-missing 0s".split(" ")).run())

                print(buf.getvalue())
                self.assertNotIn(": Destroy missing", buf.getvalue())


        with self.subTest("missing dataset of us that still has children"):

            #just deselect it so it counts as 'missing'
            shelltest("zfs set autobackup:test=child test_source1/fs1")

            with OutputIO() as buf:
                with redirect_stdout(buf), redirect_stderr(buf):
                        self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-snapshot --destroy-missing 0s".split(" ")).run())

                print(buf.getvalue())
                #should have done the snapshot cleanup for destoy missing:
                self.assertIn("fs1@test-19101111000000: Destroying", buf.getvalue())

                self.assertIn("fs1: Destroy missing: Still has children here.", buf.getvalue())

            shelltest("zfs inherit autobackup:test test_source1/fs1")


        with self.subTest("Normal destroyed leaf"):
            shelltest("zfs destroy -r test_source1/fs1/sub")

            #wait for deadline of last snapshot
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    #100y: lastest should not be old enough, while second to latest snapshot IS old enough:
                    self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-snapshot --destroy-missing 100y".split(" ")).run())

                print(buf.getvalue())
                self.assertIn(": Waiting for deadline", buf.getvalue())

            #past deadline, destroy
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-snapshot --destroy-missing 1y".split(" ")).run())

                print(buf.getvalue())
                self.assertIn("sub: Destroying", buf.getvalue())


        with self.subTest("Leaf with other snapshot still using it"):
            shelltest("zfs destroy -r test_source1/fs1")
            shelltest("zfs snapshot -r test_target1/test_source1/fs1@other1")


            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-snapshot --destroy-missing 0s".split(" ")).run())

                print(buf.getvalue())

                #cant finish because still in use:
                self.assertIn("fs1: Destroy missing: Still in use", buf.getvalue())

            shelltest("zfs destroy test_target1/test_source1/fs1@other1")


        with self.subTest("In use by clone"):
            shelltest("zfs clone test_target1/test_source1/fs1@test-20101111000000 test_target1/clone1")

            with OutputIO() as buf:
                with redirect_stdout(buf), redirect_stderr(buf):
                        self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-snapshot --destroy-missing 0s".split(" ")).run())

                print(buf.getvalue())
                #now tries to destroy our own last snapshot (before the final destroy of the dataset)
                self.assertIn("fs1@test-20101111000000: Destroying", buf.getvalue())
                #but cant finish because still in use:
                self.assertIn("fs1: Error during --destroy-missing", buf.getvalue())

            shelltest("zfs destroy test_target1/clone1")


        with self.subTest("Should leave test_source1 parent"):

            with OutputIO() as buf:
                with redirect_stdout(buf), redirect_stderr(buf):
                        self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-snapshot --destroy-missing 0s".split(" ")).run())

                print(buf.getvalue())
                #should have done the snapshot cleanup for destoy missing:
                self.assertIn("fs1: Destroying", buf.getvalue())

            with OutputIO() as buf:
                with redirect_stdout(buf), redirect_stderr(buf):
                        self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-snapshot --destroy-missing 0s".split(" ")).run())

                print(buf.getvalue())
                #on second run it sees the dangling ex-parent but doesnt know what to do with it (since it has no own snapshot)
                self.assertIn("test_source1: Destroy missing: has no snapshots made by us", buf.getvalue())




        #end result
        r=shelltest("zfs list -H -o name -r -t all test_target1")
        self.assertMultiLineEqual(r,"""
test_target1
test_target1/test_source1
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-19101111000000
test_target1/test_source2/fs2/sub@test-20101111000000
""")




from .ThinnerRule import ThinnerRule


class Thinner:
    """progressive thinner (universal, used for cleaning up snapshots)"""

    def __init__(self, schedule_str=""):
        """
        Args:
            schedule_str: comma seperated list of ThinnerRules. A plain number specifies how many snapshots to always keep.
        """

        self.rules = []
        self.always_keep = 0

        if schedule_str == "":
            return

        rule_strs = schedule_str.split(",")
        for rule_str in rule_strs:
            if rule_str.lstrip('-').isdigit():
                self.always_keep = int(rule_str)
                if self.always_keep < 0:
                    raise (Exception("Number of snapshots to keep cant be negative: {}".format(self.always_keep)))
            else:
                self.rules.append(ThinnerRule(rule_str))

    def human_rules(self):
        """get list of human readable rules"""
        ret = []
        if self.always_keep:
            ret.append("Keep the last {} snapshot{}.".format(self.always_keep, self.always_keep != 1 and "s" or ""))
        for rule in self.rules:
            ret.append(rule.human_str)

        return ret

    def thin(self, objects, keep_objects, now):
        """thin list of objects with current schedule rules. objects: list of
        objects to thin. every object should have timestamp attribute.

            return( keeps, removes )

        Args:
            objects: list of objects to check (should have a timestamp attribute)
            keep_objects: objects to always keep (if they also are in the in the normal objects list)
            now: if specified, use this time as current time
        """


        # always keep a number of the last objets?
        if self.always_keep:
            # all of them
            if len(objects) <= self.always_keep:
                return objects, []

            # determine which ones
            always_keep_objects = objects[-self.always_keep:]
        else:
            always_keep_objects = []

        # determine time blocks
        time_blocks = {}
        for rule in self.rules:
            time_blocks[rule.period] = {}

        keeps = []
        removes = []

        # traverse objects
        for thisobject in objects:
            # important they are ints!
            timestamp = int(thisobject.timestamp)
            age = int(now) - timestamp

            # store in the correct time blocks, per period-size, if not too old yet
            # e.g.: look if there is ANY timeblock that wants to keep this object
            keep = False
            for rule in self.rules:
                if age <= rule.ttl:
                    block_nr = int(timestamp / rule.period)
                    if block_nr not in time_blocks[rule.period]:
                        time_blocks[rule.period][block_nr] = True
                        keep = True

            # keep it according to schedule, or keep it because it is in the keep_objects list
            if keep or thisobject in keep_objects or thisobject in always_keep_objects:
                keeps.append(thisobject)
            else:
                removes.append(thisobject)

        return keeps, removes


import os
# To run tests as non-root, use this hack:
# chmod 4755 /usr/sbin/zpool /usr/sbin/zfs

import sys

import zfs_autobackup.util

#dirty hack for this error:
#AttributeError: module 'collections' has no attribute 'MutableMapping'

if sys.version_info.major == 3 and sys.version_info.minor >= 10:
    import collections
    setattr(collections, "MutableMapping", collections.abc.MutableMapping)

import subprocess
import random

#default test stuff
import unittest2
import subprocess
import time
from pprint import *
from zfs_autobackup.ZfsAutobackup import *
from zfs_autobackup.ZfsAutoverify import *
from zfs_autobackup.ZfsCheck import *
from zfs_autobackup.util import *
from mock import *
import contextlib
import sys
import io

import datetime


TEST_POOLS="test_source1 test_source2 test_target1"
# ZFS_USERSPACE=  subprocess.check_output("dpkg-query -W zfsutils-linux |cut -f2", shell=True).decode('utf-8').rstrip()
# ZFS_KERNEL=     subprocess.check_output("modinfo zfs|grep ^version |sed 's/.* //'", shell=True).decode('utf-8').rstrip()

print("###########################################")
print("#### Unit testing against:")
print("#### Python                : "+sys.version.replace("\n", " "))
print("#### ZFS version           : "+subprocess.check_output("zfs --version", shell=True).decode('utf-8').rstrip().replace('\n', ' '))
print("#############################################")



# for python2 compatibility
if sys.version_info.major==2:
    OutputIO=io.BytesIO
else:
    OutputIO=io.StringIO

# for when we're using a suid-root python binary during development
os.setuid(0)
os.setgid(0)


# for python2 compatibility (python 3 has this already)
@contextlib.contextmanager
def redirect_stdout(target):
    original = sys.stdout
    try:
        sys.stdout = target
        yield
    finally:
        sys.stdout = original

# for python2 compatibility (python 3 has this already)
@contextlib.contextmanager
def redirect_stderr(target):
    original = sys.stderr
    try:
        sys.stderr = target
        yield
    finally:
        sys.stderr = original



def shelltest(cmd):
    """execute and print result as nice copypastable string for unit tests (adds extra newlines on top/bottom)"""

    ret=(subprocess.check_output(cmd , shell=True).decode('utf-8'))

    print("######### result of: {}".format(cmd))
    print(ret)
    print("#########")
    ret='\n'+ret
    return(ret)

def prepare_zpools():
    print("Preparing zfs filesystems...")

    #need ram blockdevice
    # subprocess.check_call("modprobe brd rd_size=512000", shell=True)

    #remove old stuff
    subprocess.call("zpool destroy test_source1 2>/dev/null", shell=True)
    subprocess.call("zpool destroy test_source2 2>/dev/null", shell=True)
    subprocess.call("zpool destroy test_target1 2>/dev/null", shell=True)

    #create pools
    subprocess.check_call("zpool create test_source1 /dev/ram0", shell=True)
    subprocess.check_call("zpool create test_source2 /dev/ram1", shell=True)
    subprocess.check_call("zpool create test_target1 /dev/ram2", shell=True)

    #create test structure
    subprocess.check_call("zfs create -p test_source1/fs1/sub", shell=True)
    subprocess.check_call("zfs create -p test_source2/fs2/sub", shell=True)
    subprocess.check_call("zfs create -p test_source2/fs3/sub", shell=True)
    subprocess.check_call("zfs set autobackup:test=true test_source1/fs1", shell=True)
    subprocess.check_call("zfs set autobackup:test=child test_source2/fs2", shell=True)

    print("Prepare done")



@contextlib.contextmanager
def mocktime(time_str, format="%Y%m%d%H%M%S"):

    def fake_datetime_now():
        return datetime.datetime.strptime(time_str, format)

    with patch.object(zfs_autobackup.util,'datetime_now_mock', fake_datetime_now()):
        yield







from basetest import *
from zfs_autobackup.BlockHasher import BlockHasher


# sha1 sums of files, (bs=4096)
# da39a3ee5e6b4b0d3255bfef95601890afd80709  empty
# 642027d63bb0afd7e0ba197f2c66ad03e3d70de1  partial
# 3c0bf91170d873b8e327d3bafb6bc074580d11b7  whole
# 2e863f1fcccd6642e4e28453eba10d2d3f74d798  whole2
# 959e6b58078f0cfd2fb3d37e978fda51820473ff  whole_whole2
# 309ffffba2e1977d12f3b7469971f30d28b94bd8  whole_whole2_partial


class TestTreeHasher(unittest2.TestCase):

    def test_treehasher(self):
        shelltest("rm -rf /tmp/treehashertest; mkdir /tmp/treehashertest")
        shelltest("cp tests/data/whole /tmp/treehashertest")
        shelltest("mkdir /tmp/treehashertest/emptydir")
        shelltest("mkdir /tmp/treehashertest/dir")
        shelltest("cp tests/data/whole_whole2_partial /tmp/treehashertest/dir")

        # it should ignore these:
        shelltest("ln -s / /tmp/treehashertest/symlink")
        shelltest("mknod /tmp/treehashertest/c c 1 1")
        shelltest("mknod /tmp/treehashertest/b b 1 1")
        shelltest("mkfifo /tmp/treehashertest/f")


        block_hasher = BlockHasher(count=1, skip=0)
        tree_hasher = TreeHasher(block_hasher)
        with self.subTest("Test output, count 1, skip 0"):
            self.assertEqual(list(tree_hasher.generate("/tmp/treehashertest")), [
                ('whole', 0, '3c0bf91170d873b8e327d3bafb6bc074580d11b7'),
                ('dir/whole_whole2_partial', 0, '3c0bf91170d873b8e327d3bafb6bc074580d11b7'),
                ('dir/whole_whole2_partial', 1, '2e863f1fcccd6642e4e28453eba10d2d3f74d798'),
                ('dir/whole_whole2_partial', 2, '642027d63bb0afd7e0ba197f2c66ad03e3d70de1')
            ])

        block_hasher = BlockHasher(count=1, skip=1)
        tree_hasher = TreeHasher(block_hasher)
        with self.subTest("Test output, count 1, skip 1"):
            self.assertEqual(list(tree_hasher.generate("/tmp/treehashertest")), [
                ('whole', 0, '3c0bf91170d873b8e327d3bafb6bc074580d11b7'),
                # ('dir/whole_whole2_partial', 0, '3c0bf91170d873b8e327d3bafb6bc074580d11b7'),
                ('dir/whole_whole2_partial', 1, '2e863f1fcccd6642e4e28453eba10d2d3f74d798'),
                # ('dir/whole_whole2_partial', 2, '642027d63bb0afd7e0ba197f2c66ad03e3d70de1')
            ])



        block_hasher = BlockHasher(count=2)
        tree_hasher = TreeHasher(block_hasher)

        with self.subTest("Test output, count 2, skip 0"):
            self.assertEqual(list(tree_hasher.generate("/tmp/treehashertest")), [
                ('whole', 0, '3c0bf91170d873b8e327d3bafb6bc074580d11b7'),
                ('dir/whole_whole2_partial', 0, '959e6b58078f0cfd2fb3d37e978fda51820473ff'),
                ('dir/whole_whole2_partial', 1, '642027d63bb0afd7e0ba197f2c66ad03e3d70de1')
            ])

        with self.subTest("Test compare"):
            generator = tree_hasher.generate("/tmp/treehashertest")
            errors = list(tree_hasher.compare("/tmp/treehashertest", generator))
            self.assertEqual(errors, [])

        with self.subTest("Test mismatch"):
            generator = list(tree_hasher.generate("/tmp/treehashertest"))
            shelltest("cp tests/data/whole2 /tmp/treehashertest/whole")

            self.assertEqual(list(tree_hasher.compare("/tmp/treehashertest", generator)),
                             [('whole',
                               0,
                               '3c0bf91170d873b8e327d3bafb6bc074580d11b7',
                               '2e863f1fcccd6642e4e28453eba10d2d3f74d798')])

        with self.subTest("Test missing file compare"):
            generator = list(tree_hasher.generate("/tmp/treehashertest"))
            shelltest("rm /tmp/treehashertest/whole")

            self.assertEqual(list(tree_hasher.compare("/tmp/treehashertest", generator)),
                             [('whole', '-', '-', "ERROR: [Errno 2] No such file or directory: '/tmp/treehashertest/whole'")])





from os.path import exists

from basetest import *
from zfs_autobackup.BlockHasher import BlockHasher


class TestZfsCheck(unittest2.TestCase):

    def setUp(self):
        pass


    def test_volume(self):

        if exists("/.dockerenv"):
            self.skipTest("FIXME: zfscheck volumes not supported in docker yet")

        prepare_zpools()

        shelltest("zfs create -V200M test_source1/vol")
        shelltest("zfs snapshot test_source1/vol@test")

        with self.subTest("Generate"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertFalse(ZfsCheck("test_source1/vol@test".split(" "),print_arguments=False).run())

                print(buf.getvalue())
                self.assertEqual("""0	2c2ceccb5ec5574f791d45b63c940cff20550f9a
1	2c2ceccb5ec5574f791d45b63c940cff20550f9a
""", buf.getvalue())

                #store on disk for next step, add one error.
                with open("/tmp/testhashes", "w") as fh:
                    fh.write(buf.getvalue()+"1\t2c2ceccb5ec5574f791d45b63c940cff20550f9X")

        with self.subTest("Compare"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertEqual(1, ZfsCheck("test_source1/vol@test --check=/tmp/testhashes".split(" "),print_arguments=False).run())
                print(buf.getvalue())
                self.assertEqual("Chunk 1 failed: 2c2ceccb5ec5574f791d45b63c940cff20550f9X 2c2ceccb5ec5574f791d45b63c940cff20550f9a\n", buf.getvalue())

    def test_filesystem(self):
        prepare_zpools()

        shelltest("cp tests/data/whole /test_source1/testfile")
        shelltest("mkdir /test_source1/emptydir")
        shelltest("mkdir /test_source1/dir")
        shelltest("cp tests/data/whole2 /test_source1/dir/testfile")

        #it should ignore these:
        shelltest("ln -s / /test_source1/symlink")
        shelltest("mknod /test_source1/c c 1 1")
        shelltest("mknod /test_source1/b b 1 1")
        shelltest("mkfifo /test_source1/f")

        shelltest("zfs snapshot test_source1@test")
        ZfsCheck("test_source1@test --debug".split(" "), print_arguments=False).run()
        with self.subTest("Generate"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertFalse(ZfsCheck("test_source1@test".split(" "), print_arguments=False).run())

                print(buf.getvalue())
                self.assertEqual("""testfile	0	3c0bf91170d873b8e327d3bafb6bc074580d11b7
dir/testfile	0	2e863f1fcccd6642e4e28453eba10d2d3f74d798
""", buf.getvalue())

                #store on disk for next step, add error
                with open("/tmp/testhashes", "w") as fh:
                    fh.write(buf.getvalue()+"dir/testfile	0	2e863f1fcccd6642e4e28453eba10d2d3f74d79X")

        with self.subTest("Compare"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertEqual(1, ZfsCheck("test_source1@test --check=/tmp/testhashes".split(" "),print_arguments=False).run())

                print(buf.getvalue())
                self.assertEqual("dir/testfile: Chunk 0 failed: 2e863f1fcccd6642e4e28453eba10d2d3f74d79X 2e863f1fcccd6642e4e28453eba10d2d3f74d798\n", buf.getvalue())

    def test_file(self):

        with self.subTest("Generate"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertFalse(ZfsCheck("tests/data/whole".split(" "), print_arguments=False).run())

                print(buf.getvalue())
                self.assertEqual("""0	3c0bf91170d873b8e327d3bafb6bc074580d11b7
""", buf.getvalue())

                # store on disk for next step, add error
                with open("/tmp/testhashes", "w") as fh:
                    fh.write(buf.getvalue()+"0	3c0bf91170d873b8e327d3bafb6bc074580d11bX")

        with self.subTest("Compare"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertEqual(1,ZfsCheck("tests/data/whole --check=/tmp/testhashes".split(" "), print_arguments=False).run())
                print(buf.getvalue())
                self.assertEqual("Chunk 0 failed: 3c0bf91170d873b8e327d3bafb6bc074580d11bX 3c0bf91170d873b8e327d3bafb6bc074580d11b7\n", buf.getvalue())

    def test_tree(self):
        shelltest("rm -rf /tmp/testtree; mkdir /tmp/testtree")
        shelltest("cp tests/data/whole /tmp/testtree")
        shelltest("cp tests/data/whole_whole2 /tmp/testtree")
        shelltest("cp tests/data/whole2 /tmp/testtree")
        shelltest("cp tests/data/partial /tmp/testtree")
        shelltest("cp tests/data/whole_whole2_partial /tmp/testtree")

        ####################################
        with self.subTest("Generate, skip 1"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertFalse(ZfsCheck("/tmp/testtree --skip=1".split(" "), print_arguments=False).run())

                #since order varies, just check count (there is one empty line for some reason, only when testing like this)
                print(buf.getvalue().split("\n"))
                self.assertEqual(len(buf.getvalue().split("\n")),4)

        ######################################
        with self.subTest("Compare, all incorrect, skip 1"):

            # store on disk for next step, add error
            with open("/tmp/testhashes", "w") as fh:
                fh.write("""
partial	0	642027d63bb0afd7e0ba197f2c66ad03e3d70deX
whole	0	3c0bf91170d873b8e327d3bafb6bc074580d11bX
whole2	0	2e863f1fcccd6642e4e28453eba10d2d3f74d79X
whole_whole2	0	959e6b58078f0cfd2fb3d37e978fda51820473fX
whole_whole2_partial	0	309ffffba2e1977d12f3b7469971f30d28b94bdX
""")

            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertEqual(ZfsCheck("/tmp/testtree --check=/tmp/testhashes --skip=1".split(" "), print_arguments=False).run(), 3)

                print(buf.getvalue())
                self.assertMultiLineEqual("""partial: Chunk 0 failed: 642027d63bb0afd7e0ba197f2c66ad03e3d70deX 642027d63bb0afd7e0ba197f2c66ad03e3d70de1
whole2: Chunk 0 failed: 2e863f1fcccd6642e4e28453eba10d2d3f74d79X 2e863f1fcccd6642e4e28453eba10d2d3f74d798
whole_whole2_partial: Chunk 0 failed: 309ffffba2e1977d12f3b7469971f30d28b94bdX 309ffffba2e1977d12f3b7469971f30d28b94bd8
""",buf.getvalue())

        ####################################
        with self.subTest("Generate"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertFalse(ZfsCheck("/tmp/testtree".split(" "), print_arguments=False).run())

                #file order on disk can vary, so sort it..
                sorted=buf.getvalue().split("\n")
                sorted.sort()
                sorted="\n".join(sorted)+"\n"

                print(sorted)
                self.assertEqual("""
partial	0	642027d63bb0afd7e0ba197f2c66ad03e3d70de1
whole	0	3c0bf91170d873b8e327d3bafb6bc074580d11b7
whole2	0	2e863f1fcccd6642e4e28453eba10d2d3f74d798
whole_whole2	0	959e6b58078f0cfd2fb3d37e978fda51820473ff
whole_whole2_partial	0	309ffffba2e1977d12f3b7469971f30d28b94bd8
""", sorted)

                # store on disk for next step, add error
                with open("/tmp/testhashes", "w") as fh:
                    fh.write(buf.getvalue() + "whole_whole2_partial	0	309ffffba2e1977d12f3b7469971f30d28b94bdX")

        ####################################
        with self.subTest("Compare"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    self.assertEqual(1, ZfsCheck("/tmp/testtree --check=/tmp/testhashes".split(" "),
                                                 print_arguments=False).run())
                print(buf.getvalue())
                self.assertEqual(
                    "whole_whole2_partial: Chunk 0 failed: 309ffffba2e1977d12f3b7469971f30d28b94bdX 309ffffba2e1977d12f3b7469971f30d28b94bd8\n",
                    buf.getvalue())

    def test_brokenpipe_cleanup_filesystem(self):
        """test if stuff is cleaned up correctly, in debugging mode , when a pipe breaks. """

        prepare_zpools()
        shelltest("cp tests/data/whole /test_source1/testfile")
        shelltest("zfs snapshot test_source1@test")

        #breaks pipe when head exists
        #important to use --debug, since that generates extra output which would be problematic if we didnt do correct SIGPIPE handling
        shelltest("python -m zfs_autobackup.ZfsCheck test_source1@test --debug | head -n1")

        #should NOT be mounted anymore if cleanup went ok:
        self.assertNotRegex(shelltest("mount"), "test_source1@test")

    def test_brokenpipe_cleanup_volume(self):
        if exists("/.dockerenv"):
            self.skipTest("FIXME: zfscheck volumes not supported in docker yet")

        prepare_zpools()
        shelltest("zfs create -V200M test_source1/vol")
        shelltest("zfs snapshot test_source1/vol@test")

        #breaks pipe when grep exists:
        #important to use --debug, since that generates extra output which would be problematic if we didnt do correct SIGPIPE handling
        shelltest("python -m zfs_autobackup.ZfsCheck test_source1/vol@test --debug| grep -m1 'Hashing file'")
        # time.sleep(1)

        r = shelltest("zfs list -H -o name -r -t all " + TEST_POOLS)
        self.assertMultiLineEqual("""
test_source1
test_source1/fs1
test_source1/fs1/sub
test_source1/vol
test_source1/vol@test
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs3
test_source2/fs3/sub
test_target1
""",r )






import argparse
import os.path
import sys

from .LogConsole import LogConsole


class CliBase(object):
    """Base class for all cli programs
    Overridden in subclasses that add stuff for the specific programs."""

    # also used by setup.py
    VERSION = "3.3"
    HEADER = "{} v{} - (c)2022 E.H.Eefting (edwin@datux.nl)".format(os.path.basename(sys.argv[0]), VERSION)

    def __init__(self, argv, print_arguments=True):

        self.parser=self.get_parser()
        self.args = self.parse_args(argv)

        # helps with investigating failed regression tests:
        if print_arguments:
            print("ARGUMENTS: " + " ".join(argv))

    def parse_args(self, argv):
        """parses the arguments and does additional checks, might print warnings or notes
        Overridden in subclasses with extra checks.
        """

        args = self.parser.parse_args(argv)

        if args.help:
            self.parser.print_help()
            sys.exit(255)

        if args.version:
            print(self.HEADER)
            sys.exit(255)

        # auto enable progress?
        if sys.stderr.isatty() and not args.no_progress:
            args.progress = True

        if args.debug_output:
            args.debug = True

        if args.test:
            args.verbose = True

        if args.debug:
            args.verbose = True

        self.log = LogConsole(show_debug=args.debug, show_verbose=args.verbose, color=sys.stdout.isatty())

        self.verbose(self.HEADER)
        self.verbose("")

        return args

    def get_parser(self):
        """build up the argument parser
        Overridden in subclasses that add extra arguments
        """

        parser = argparse.ArgumentParser(description=self.HEADER, add_help=False,
                                         epilog='Full manual at: https://github.com/psy0rz/zfs_autobackup')

        # Basic options
        group=parser.add_argument_group("Common options")
        group.add_argument('--help', '-h', action='store_true', help='show help')
        group.add_argument('--test', '--dry-run', '-n', action='store_true',
                            help='Dry run, dont change anything, just show what would be done (still does all read-only '
                                 'operations)')
        group.add_argument('--verbose', '-v', action='store_true', help='verbose output')
        group.add_argument('--debug', '-d', action='store_true',
                            help='Show zfs commands that are executed, stops after an exception.')
        group.add_argument('--debug-output', action='store_true',
                            help='Show zfs commands and their output/exit codes. (noisy)')
        group.add_argument('--progress', action='store_true',
                            help='show zfs progress output. Enabled automaticly on ttys. (use --no-progress to disable)')
        group.add_argument('--no-progress', action='store_true',
                            help=argparse.SUPPRESS)  # needed to workaround a zfs recv -v bug
        group.add_argument('--utc', action='store_true',
                            help='Use UTC instead of local time when dealing with timestamps for both formatting and parsing. To snapshot in an ISO 8601 compliant time format you may for example specify --snapshot-format "{}-%%Y-%%m-%%dT%%H:%%M:%%SZ". Changing this parameter after-the-fact (existing snapshots) will cause their timestamps to be interpreted as a different time than before.')
        group.add_argument('--version', action='store_true',
                            help='Show version.')


        return parser

    def verbose(self, txt):
        self.log.verbose(txt)

    def warning(self, txt):
        self.log.warning(txt)

    def error(self, txt):
        self.log.error(txt)

    def debug(self, txt):
        self.log.debug(txt)

    def progress(self, txt):
        self.log.progress(txt)

    def clear_progress(self):
        self.log.clear_progress()

    def set_title(self, title):
        self.log.verbose("")
        self.log.verbose("#### " + title)



# python 2 compatibility
from __future__ import print_function

import sys

class LogConsole:
    """Log-class that outputs to console, adding colors if needed"""

    def __init__(self, show_debug, show_verbose, color):
        self.last_log = ""
        self.show_debug = show_debug
        self.show_verbose = show_verbose
        self._progress_uncleared=False

        if color:
            # try to use color, failback if colorama not available
            self.colorama=False
            try:
                import colorama
                global colorama
                self.colorama = True
            except ImportError:
                pass

        else:
            self.colorama=False

    def error(self, txt):
        self.clear_progress()
        if self.colorama:
            print(colorama.Fore.RED + colorama.Style.BRIGHT + "! " + txt + colorama.Style.RESET_ALL, file=sys.stderr)
        else:
            print("! " + txt, file=sys.stderr)
        sys.stderr.flush()

    def warning(self, txt):
        self.clear_progress()
        if self.colorama:
            print(colorama.Fore.YELLOW + colorama.Style.NORMAL + "  NOTE: " + txt + colorama.Style.RESET_ALL)
        else:
            print("  NOTE: " + txt)
        sys.stdout.flush()

    def verbose(self, txt):
        if self.show_verbose:
            self.clear_progress()
            if self.colorama:
                print(colorama.Style.NORMAL + "  " + txt + colorama.Style.RESET_ALL)
            else:
                print("  " + txt)
            sys.stdout.flush()

    def debug(self, txt):
        if self.show_debug:
            self.clear_progress()
            if self.colorama:
                print(colorama.Fore.GREEN + "# " + txt + colorama.Style.RESET_ALL)
            else:
                print("# " + txt)
            sys.stdout.flush()

    def progress(self, txt):
        """print progress output to stderr (stays on same line)"""
        self.clear_progress()
        self._progress_uncleared=True
        print(">>> {}\r".format(txt), end='', file=sys.stderr)
        sys.stderr.flush()

    def clear_progress(self):
        if self._progress_uncleared:
            import colorama
            print(colorama.ansi.clear_line(), end='', file=sys.stderr)
            # sys.stderr.flush()
            self._progress_uncleared=False



import re


class ThinnerRule:
    """a thinning schedule rule for Thinner"""

    TIME_NAMES = {
        'y': 3600 * 24 * 365.25,
        'm': 3600 * 24 * 30,
        'w': 3600 * 24 * 7,
        'd': 3600 * 24,
        'h': 3600,
        'min': 60,
        's': 1,
    }

    TIME_DESC = {
        'y': 'year',
        'm': 'month',
        'w': 'week',
        'd': 'day',
        'h': 'hour',
        'min': 'minute',
        's': 'second',
    }

    def __init__(self, rule_str):
        """parse scheduling string
            example:
                daily snapshot, remove after a week:     1d1w
                weekly snapshot, remove after a month:   1w1m
                monthly snapshot, remove after 6 months: 1m6m
                yearly snapshot, remove after 2 year:    1y2y
                keep all snapshots, remove after a day   1s1d
                keep nothing:                            1s1s

        """

        rule_str = rule_str.lower()
        matches = re.findall("([0-9]*)([a-z]*)([0-9]*)([a-z]*)", rule_str)[0]

        if '' in matches:
            raise (Exception("Invalid schedule string: '{}'".format(rule_str)))

        period_amount = int(matches[0])
        period_unit = matches[1]
        ttl_amount = int(matches[2])
        ttl_unit = matches[3]

        if period_unit not in self.TIME_NAMES:
            raise (Exception("Invalid period string in schedule: '{}'".format(rule_str)))

        if ttl_unit not in self.TIME_NAMES:
            raise (Exception("Invalid ttl string in schedule: '{}'".format(rule_str)))

        self.period = period_amount * self.TIME_NAMES[period_unit]
        self.ttl = ttl_amount * self.TIME_NAMES[ttl_unit]

        if self.period > self.ttl:
            raise (Exception("Period cant be longer than ttl in schedule: '{}'".format(rule_str)))

        self.rule_str = rule_str

        self.human_str = "Keep every {} {}{}, delete after {} {}{}.".format(
            period_amount, self.TIME_DESC[period_unit], period_amount != 1 and "s" or "", ttl_amount,
            self.TIME_DESC[ttl_unit], ttl_amount != 1 and "s" or "")

    def __str__(self):
        """get schedule as a schedule string"""

        return self.rule_str


from zfs_autobackup.LogConsole import LogConsole
from basetest import *


class TestLog(unittest2.TestCase):

    def test_colored(self):
        """test with color output"""
        with OutputIO() as buf:
            with redirect_stdout(buf):
                l= LogConsole(show_verbose=False, show_debug=False, color=True)
                l.verbose("verbose")
                l.debug("debug")

            with redirect_stdout(buf):
                l=LogConsole(show_verbose=True, show_debug=True, color=True)
                l.verbose("verbose")
                l.debug("debug")

            with redirect_stderr(buf):
                l=LogConsole(show_verbose=False, show_debug=False, color=True)
                l.error("error")

            print(list(buf.getvalue()))
            self.assertEqual(list(buf.getvalue()), ['\x1b', '[', '2', '2', 'm', ' ', ' ', 'v', 'e', 'r', 'b', 'o', 's', 'e', '\x1b', '[', '0', 'm', '\n', '\x1b', '[', '3', '2', 'm', '#', ' ', 'd', 'e', 'b', 'u', 'g', '\x1b', '[', '0', 'm', '\n', '\x1b', '[', '3', '1', 'm', '\x1b', '[', '1', 'm', '!', ' ', 'e', 'r', 'r', 'o', 'r', '\x1b', '[', '0', 'm', '\n'])

    def test_nocolor(self):
        """test without color output"""

        with OutputIO() as buf:
            with redirect_stdout(buf):
                l=LogConsole(show_verbose=False, show_debug=False, color=False)
                l.verbose("verbose")
                l.debug("debug")

            with redirect_stdout(buf):
                l=LogConsole(show_verbose=True, show_debug=True, color=False)
                l.verbose("verbose")
                l.debug("debug")

            with redirect_stderr(buf):
                l=LogConsole(show_verbose=False, show_debug=False, color=False)
                l.error("error")

            print(list(buf.getvalue()))
            self.assertEqual(list(buf.getvalue()), [' ', ' ', 'v', 'e', 'r', 'b', 'o', 's', 'e', '\n', '#', ' ', 'd', 'e', 'b', 'u', 'g', '\n', '!', ' ', 'e', 'r', 'r', 'o', 'r', '\n'])


        # zfs_autobackup.LogConsole.colorama=False






# from util import activate_volume_snapshot, create_mountpoints, cleanup_mountpoint
from signal import signal, SIGPIPE
from .util import output_redir, sigpipe_handler

from .ZfsAuto import ZfsAuto
from .ZfsNode import ZfsNode
import sys


# # try to be as unix compatible as possible, while still having decent performance
# def compare_trees_find(source_node, source_path, target_node, target_path):
#     # find /tmp/zfstmp_pve1_1993135target/ -xdev -type f -print0 | xargs -0 md5sum | md5sum -c
#
#     #verify tree has atleast one file
#
#     stdout=source_node.run(["find", ".", "-type", "f",
#                           ExecuteNode.PIPE, "head", "-n1",
#                           ], cwd=source_path)
#
#     if not stdout:
#         source_node.debug("No files, skipping check")
#     else:
#         pipe=source_node.run(["find", ".", "-type", "f", "-print0",
#                               ExecuteNode.PIPE, "xargs", "-0", "md5sum"
#                               ], pipe=True, cwd=source_path)
#         stdout=target_node.run([ "md5sum", "-c", "--quiet"], inp=pipe, cwd=target_path, valid_exitcodes=[0,1])
#
#         if len(stdout):
#             for line in stdout:
#                 target_node.error("md5sum: "+line)
#
#             raise(Exception("Some files have checksum errors"))
#
#
# def compare_trees_rsync(source_node, source_path, target_node, target_path):
#     """use rsync to compare two trees.
#      Advantage is that we can see which individual files differ.
#      But requires rsync and cant do remote to remote."""
#
#     cmd = ["rsync", "-rcnq", "--info=COPY,DEL,MISC,NAME,SYMSAFE", "--msgs2stderr", "--delete" ]
#
#     #local
#     if source_node.ssh_to is None and target_node.ssh_to is None:
#         cmd.append("{}/".format(source_path))
#         cmd.append("{}/".format(target_path))
#         source_node.debug("Running rsync locally, on source.")
#         stdout, stderr = source_node.run(cmd, return_stderr=True)
#
#     #source is local
#     elif source_node.ssh_to is None and target_node.ssh_to is not None:
#         cmd.append("{}/".format(source_path))
#         cmd.append("{}:{}/".format(target_node.ssh_to, target_path))
#         source_node.debug("Running rsync locally, on source.")
#         stdout, stderr = source_node.run(cmd, return_stderr=True)
#
#     #target is local
#     elif source_node.ssh_to is not None and target_node.ssh_to is None:
#         cmd.append("{}:{}/".format(source_node.ssh_to, source_path))
#         cmd.append("{}/".format(target_path))
#         source_node.debug("Running rsync locally, on target.")
#         stdout, stderr=target_node.run(cmd, return_stderr=True)
#
#     else:
#         raise Exception("Source and target cant both be remote when verifying. (rsync limitation)")
#
#     if stderr:
#         raise Exception("Dataset verify failed, see above list for differences")


def verify_filesystem(source_snapshot, source_mnt, target_snapshot, target_mnt, method):
    """Compare the contents of two zfs filesystem snapshots """

    try:

        # mount the snapshots
        source_snapshot.mount(source_mnt)
        target_snapshot.mount(target_mnt)

        if method=='rsync':
            compare_trees_rsync(source_snapshot.zfs_node, source_mnt, target_snapshot.zfs_node, target_mnt)
        # elif method == 'tar':
        #     compare_trees_tar(source_snapshot.zfs_node, source_mnt, target_snapshot.zfs_node, target_mnt)
        elif method == 'find':
            compare_trees_find(source_snapshot.zfs_node, source_mnt, target_snapshot.zfs_node, target_mnt)
        else:
            raise(Exception("program errror, unknown method"))

    finally:
        source_snapshot.unmount(source_mnt)
        target_snapshot.unmount(target_mnt)


# def hash_dev(node, dev):
#     """calculate md5sum of a device on a node"""
#
#     node.debug("Hashing volume {} ".format(dev))
#
#     cmd = [ "md5sum", dev ]
#
#     stdout = node.run(cmd)
#
#     if node.readonly:
#         hashed=None
#     else:
#         hashed = stdout[0].split(" ")[0]
#
#     node.debug("Hash of volume {} is {}".format(dev, hashed))
#
#     return hashed



# def deacitvate_volume_snapshot(snapshot):
#     clone_name=get_tmp_clone_name(snapshot)
#     clone=snapshot.zfs_node.get_dataset(clone_name)
#     clone.destroy(deferred=True, verbose=False)

def verify_volume(source_dataset, source_snapshot, target_dataset, target_snapshot):
    """compare the contents of two zfs volume snapshots"""

    # try:
    source_dev= activate_volume_snapshot(source_snapshot)
    target_dev= activate_volume_snapshot(target_snapshot)

    source_hash= hash_dev(source_snapshot.zfs_node, source_dev)
    target_hash= hash_dev(target_snapshot.zfs_node, target_dev)

    if source_hash!=target_hash:
        raise Exception("md5hash difference: {} != {}".format(source_hash, target_hash))

    # finally:
    #     deacitvate_volume_snapshot(source_snapshot)
    #     deacitvate_volume_snapshot(target_snapshot)


# class ZfsAutoChecksumVolume(ZfsAuto):
#     def __init__(self, argv, print_arguments=True):
#
#         # NOTE: common options and parameters are in ZfsAuto
#         super(ZfsAutoverify, self).__init__(argv, print_arguments)

class ZfsAutoverify(ZfsAuto):
    """The zfs-autoverify class, default agruments and stuff come from ZfsAuto"""

    def __init__(self, argv, print_arguments=True):

        # NOTE: common options and parameters are in ZfsAuto
        super(ZfsAutoverify, self).__init__(argv, print_arguments)

    def parse_args(self, argv):
        """do extra checks on common args"""

        args=super(ZfsAutoverify, self).parse_args(argv)

        if args.target_path == None:
            self.log.error("Please specify TARGET-PATH")
            sys.exit(255)

        return args

    def get_parser(self):
        """extend common parser with  extra stuff needed for zfs-autobackup"""

        parser=super(ZfsAutoverify, self).get_parser()

        group=parser.add_argument_group("Verify options")
        group.add_argument('--fs-compare', metavar='METHOD', default="find", choices=["find", "rsync"],
                            help='Compare method to use for filesystems. (find, rsync) Default: %(default)s ')

        return parser

    def verify_datasets(self, source_mnt, source_datasets, target_node, target_mnt):

        fail_count=0
        count = 0
        for source_dataset in source_datasets:

            # stats
            if self.args.progress:
                count = count + 1
                self.progress("Analysing dataset {}/{} ({} failed)".format(count, len(source_datasets), fail_count))

            try:
                # determine corresponding target_dataset
                target_name = self.make_target_name(source_dataset)
                target_dataset = target_node.get_dataset(target_name)

                # find common snapshots to  verify
                source_snapshot = source_dataset.find_common_snapshot(target_dataset, True)
                target_snapshot = target_dataset.find_snapshot(source_snapshot)

                if source_snapshot is None or target_snapshot is None:
                    raise(Exception("Cant find common snapshot"))

                target_snapshot.verbose("Verifying...")

                if source_dataset.properties['type']=="filesystem":
                    verify_filesystem(source_snapshot, source_mnt, target_snapshot, target_mnt, self.args.fs_compare)
                elif source_dataset.properties['type']=="volume":
                    verify_volume(source_dataset, source_snapshot, target_dataset, target_snapshot)
                else:
                    raise(Exception("{} has unknown type {}".format(source_dataset, source_dataset.properties['type'])))


            except Exception as e:
                # if self.args.progress:
                #     self.clear_progress()

                fail_count = fail_count + 1
                target_dataset.error("FAILED: " + str(e))
                if self.args.debug:
                    self.verbose("Debug mode, aborting on first error")
                    raise

        # if self.args.progress:
        #     self.clear_progress()

        return fail_count

    def run(self):

        source_node=None
        source_mnt=None
        target_node=None
        target_mnt=None


        try:

            ################ create source zfsNode
            self.set_title("Source settings")

            description = "[Source]"
            source_node = ZfsNode(utc=self.args.utc,
                                  snapshot_time_format=self.snapshot_time_format, hold_name=self.hold_name, logger=self,
                                  ssh_config=self.args.ssh_config,
                                  ssh_to=self.args.ssh_source, readonly=self.args.test,
                                  debug_output=self.args.debug_output, description=description,
                                  exclude_snapshot_patterns=self.args.exclude_snapshot_pattern)

            ################# select source datasets
            self.set_title("Selecting")
            ( source_datasets, excluded_datasets) = source_node.selected_datasets(property_name=self.property_name,
                                                            exclude_received=self.args.exclude_received,
                                                            exclude_paths=self.exclude_paths,
                                                            exclude_unchanged=self.args.exclude_unchanged)
            if not source_datasets and not excluded_datasets:
                self.print_error_sources()
                return 255

            # create target_node
            self.set_title("Target settings")
            target_node = ZfsNode(utc=self.args.utc,
                                  snapshot_time_format=self.snapshot_time_format, hold_name=self.hold_name,
                                  logger=self, ssh_config=self.args.ssh_config,
                                  ssh_to=self.args.ssh_target,
                                  readonly=self.args.test, debug_output=self.args.debug_output,
                                  description="[Target]")
            target_node.verbose("Verify datasets under: {}".format(self.args.target_path))

            self.set_title("Verifying")

            source_mnt, target_mnt= create_mountpoints(source_node, target_node)

            fail_count = self.verify_datasets(
                source_mnt=source_mnt,
                source_datasets=source_datasets,
                target_mnt=target_mnt,
                target_node=target_node)

            if not fail_count:
                if self.args.test:
                    self.set_title("All tests successful.")
                else:
                    self.set_title("All datasets verified ok")

            else:
                if fail_count != 255:
                    self.error("{} dataset(s) failed!".format(fail_count))

            if self.args.test:
                self.verbose("")
                self.warning("TEST MODE - DID NOT VERIFY ANYTHING!")

            return fail_count

        except Exception as e:
            self.error("Exception: " + str(e))
            if self.args.debug:
                raise
            return 255
        except KeyboardInterrupt:
            self.error("Aborted")
            return 255
        finally:

            # cleanup
            if source_mnt is not None:
                cleanup_mountpoint(source_node, source_mnt)

            if target_mnt is not None:
                cleanup_mountpoint(target_node, target_mnt)




def cli():
    import sys

    raise(Exception("This program is incomplete, dont use it yet."))
    signal(SIGPIPE, sigpipe_handler)
    failed = ZfsAutoverify(sys.argv[1:], False).run()
    sys.exit(min(failed,255))


if __name__ == "__main__":
    cli()



import re
from datetime import datetime
import sys
import time

from .CachedProperty import CachedProperty
from .ExecuteNode import ExecuteError


class ZfsDataset:
    """a zfs dataset (filesystem/volume/snapshot/clone) Note that a dataset
    doesn't have to actually exist (yet/anymore) Also most properties are cached
    for performance-reasons, but also to allow --test to function correctly.
    """

    # illegal properties per dataset type. these will be removed from --set-properties and --filter-properties
    ILLEGAL_PROPERTIES = {
        'filesystem': [],
        'volume': ["canmount"],
    }

    def __init__(self, zfs_node, name, force_exists=None):
        """
        Args:
            :type zfs_node: ZfsNode.ZfsNode
            :type name: str
            :type force_exists: bool
        """
        self.zfs_node = zfs_node
        self.name = name  # full name
        self._virtual_snapshots = []
        self.invalidate()
        self.force_exists = force_exists

    def __repr__(self):
        return "{}: {}".format(self.zfs_node, self.name)

    def __str__(self):
        return self.name

    def __eq__(self, obj):
        if not isinstance(obj, ZfsDataset):
            return False

        return self.name == obj.name

    def verbose(self, txt):
        """
        Args:
            :type txt: str
        """
        self.zfs_node.verbose("{}: {}".format(self.name, txt))

    def error(self, txt):
        """
        Args:
            :type txt: str
        """
        self.zfs_node.error("{}: {}".format(self.name, txt))

    def warning(self, txt):
        """
        Args:
            :type txt: str
        """
        self.zfs_node.warning("{}: {}".format(self.name, txt))

    def debug(self, txt):
        """
        Args:
            :type txt: str
        """
        self.zfs_node.debug("{}: {}".format(self.name, txt))

    def invalidate(self):
        """clear caches"""
        CachedProperty.clear(self)
        self.force_exists = None
        self._virtual_snapshots = []

    def split_path(self):
        """return the path elements as an array"""
        return self.name.split("/")

    def lstrip_path(self, count):
        """return name with first count components stripped

        Args:
            :type count: int
        """
        components = self.split_path()
        if count > len(components):
            raise Exception("Trying to strip too much from path ({} items from {})".format(count, self.name))

        return "/".join(components[count:])

    def rstrip_path(self, count):
        """return name with last count components stripped

        Args:
            :type count: int
        """
        return "/".join(self.split_path()[:-count])

    @property
    def filesystem_name(self):
        """filesystem part of the name (before the @)"""
        if self.is_snapshot:
            (filesystem, snapshot) = self.name.split("@")
            return filesystem
        else:
            return self.name

    @property
    def snapshot_name(self):
        """snapshot part of the name"""
        if not self.is_snapshot:
            raise (Exception("This is not a snapshot"))

        (filesystem, snapshot_name) = self.name.split("@")
        return snapshot_name

    @property
    def is_snapshot(self):
        """true if this dataset is a snapshot"""
        return self.name.find("@") != -1
    
    @property
    def is_excluded(self):
        """true if this dataset is a snapshot and matches the exclude pattern"""
        if not self.is_snapshot:
            return False
        

        for pattern in self.zfs_node.exclude_snapshot_patterns:
            if pattern.search(self.name) is not None:
                self.debug("Excluded (path matches snapshot exclude pattern)")
                return True
        
        

    def is_selected(self, value, source, inherited, exclude_received, exclude_paths, exclude_unchanged):
        """determine if dataset should be selected for backup (called from
        ZfsNode)

        Args:
            :type exclude_paths: list[str]
            :type value: str
            :type source: str
            :type inherited: bool
            :type exclude_received: bool
            :type exclude_unchanged: int

            :param value: Value of the zfs property ("false"/"true"/"child"/parent/"-")
            :param source: Source of the zfs property ("local"/"received", "-")
            :param inherited: True of the value/source was inherited from a higher dataset.

        Returns: True : Selected
                 False: Excluded
                 None: No property found
        """

        # sanity checks
        if source not in ["local", "received", "-"]:
            # probably a program error in zfs-autobackup or new feature in zfs
            raise (Exception(
                "{} autobackup-property has illegal source: '{}' (possible BUG)".format(self.name, source)))

        if value not in ["false", "true", "child", "parent", "-"]:
            # user error
            raise (Exception(
                "{} autobackup-property has illegal value: '{}'".format(self.name, value)))

        # non specified, ignore
        if value == "-":
            return None

        # only select childs of this dataset, ignore
        if value == "child" and not inherited:
            return False

        # only select parent, no childs, ignore
        if value == "parent" and inherited:
            return False

        # manually excluded by property
        if value == "false":
            self.verbose("Excluded")
            return False

        # from here on the dataset is selected by property, now do additional exclusion checks

        # our path starts with one of the excluded paths?
        for exclude_path in exclude_paths:
            # if self.name.startswith(exclude_path):
            if (self.name + "/").startswith(exclude_path + "/"):
                # too noisy for verbose
                self.debug("Excluded (path in exclude list)")
                return False

        if source == "received":
            if exclude_received:
                self.verbose("Excluded (dataset already received)")
                return False

        if not self.is_changed(exclude_unchanged):
            self.verbose("Excluded (by --exclude-unchanged)")
            return False

        self.verbose("Selected")
        return True

    @property
    def parent(self):
        """get zfs-parent of this dataset. for snapshots this means it will get
        the filesystem/volume that it belongs to. otherwise it will return the
        parent according to path

        we cache this so everything in the parent that is cached also stays.

        returns None if there is no parent.
        :rtype: ZfsDataset | None
        """
        if self.is_snapshot:
            return self.zfs_node.get_dataset(self.filesystem_name)
        else:
            stripped = self.rstrip_path(1)
            if stripped:
                return self.zfs_node.get_dataset(stripped)
            else:
                return None

    # NOTE: unused for now
    # def find_prev_snapshot(self, snapshot, also_other_snapshots=False):
    #     """find previous snapshot in this dataset. None if it doesn't exist.
    #
    #     also_other_snapshots: set to true to also return snapshots that where
    #     not created by us. (is_ours)
    #
    #     Args:
    #         :type snapshot: str or ZfsDataset.ZfsDataset
    #         :type also_other_snapshots: bool
    #     """
    #
    #     if self.is_snapshot:
    #         raise (Exception("Please call this on a dataset."))
    #
    #     index = self.find_snapshot_index(snapshot)
    #     while index:
    #         index = index - 1
    #         if also_other_snapshots or self.snapshots[index].is_ours():
    #             return self.snapshots[index]
    #     return None

    def find_next_snapshot(self, snapshot, also_other_snapshots=False):
        """find next snapshot in this dataset. None if it doesn't exist

        Args:
            :type snapshot: ZfsDataset
            :type also_other_snapshots: bool
        """

        if self.is_snapshot:
            raise (Exception("Please call this on a dataset."))

        index = self.find_snapshot_index(snapshot)
        while index is not None and index < len(self.snapshots) - 1:
            index = index + 1
            if also_other_snapshots or self.snapshots[index].is_ours():
                return self.snapshots[index]
        return None

    @CachedProperty
    def exists_check(self):
        """check on disk if it exists"""
        self.debug("Checking if dataset exists")
        return (len(self.zfs_node.run(tab_split=True, cmd=["zfs", "list", self.name], readonly=True,
                                      valid_exitcodes=[0, 1],
                                      hide_errors=True)) > 0)

    @property
    def exists(self):
        """returns True if dataset should exist.
           Use force_exists to force a specific value, if you already know. Useful for performance and test reasons
        """

        if self.force_exists is not None:
            if self.force_exists:
                self.debug("Dataset should exist")
            else:
                self.debug("Dataset should not exist")
            return self.force_exists
        else:
            return self.exists_check

    def create_filesystem(self, parents=False, unmountable=True):
        """create a filesystem

        Args:
            :type parents: bool
        """

        # recurse up
        if parents and self.parent and not self.parent.exists:
            self.parent.create_filesystem(parents, unmountable)

        cmd = ["zfs", "create"]

        if unmountable:
            cmd.extend(["-o", "canmount=off"])

        cmd.append(self.name)
        self.zfs_node.run(cmd)

        self.force_exists = True

    def destroy(self, fail_exception=False, deferred=False, verbose=True):
        """destroy the dataset. by default failures are not an exception, so we
        can continue making backups

        Args:
            :type fail_exception: bool
        """

        if verbose:
            self.verbose("Destroying")
        else:
            self.debug("Destroying")

        if self.is_snapshot:
            self.release()

        try:
            if deferred and self.is_snapshot:
                self.zfs_node.run(["zfs", "destroy", "-d", self.name])
            else:
                self.zfs_node.run(["zfs", "destroy", self.name])

            self.invalidate()
            self.force_exists = False
            return True
        except ExecuteError:
            if not fail_exception:
                return False
            else:
                raise

    @CachedProperty
    def properties(self):
        """all zfs properties"""

        cmd = [
            "zfs", "get", "-H", "-o", "property,value", "-p", "all", self.name
        ]

        self.debug("Getting zfs properties")

        ret = {}
        for pair in self.zfs_node.run(tab_split=True, cmd=cmd, readonly=True, valid_exitcodes=[0]):
            if len(pair) == 2:
                ret[pair[0]] = pair[1]

        return ret

    def is_changed(self, min_changed_bytes=1):
        """dataset is changed since ANY latest snapshot ?

        Args:
            :type min_changed_bytes: int
        """
        self.debug("Checking if dataset is changed")

        if min_changed_bytes == 0:
            return True

        if int(self.properties['written']) < min_changed_bytes:
            return False
        else:
            return True

    def is_ours(self):
        """return true if this snapshot name has format"""
        try:
            test = self.timestamp
        except ValueError as e:
            return False

        return True

    @property
    def holds(self):
        """get list[holds] for dataset"""

        output = self.zfs_node.run(["zfs", "holds", "-H", self.name], valid_exitcodes=[0], tab_split=True,
                                   readonly=True)
        return map(lambda fields: fields[1], output)

    def is_hold(self):
        """did we hold this snapshot?"""
        return self.zfs_node.hold_name in self.holds

    def hold(self):
        """hold dataset"""
        self.debug("holding")
        self.zfs_node.run(["zfs", "hold", self.zfs_node.hold_name, self.name], valid_exitcodes=[0, 1])

    def release(self):
        """release dataset"""
        if self.zfs_node.readonly or self.is_hold():
            self.debug("releasing")
            self.zfs_node.run(["zfs", "release", self.zfs_node.hold_name, self.name], valid_exitcodes=[0, 1])

    @property
    def timestamp(self):
        """get timestamp from snapshot name. Only works for our own snapshots
        with the correct format.
        """
        dt = datetime.strptime(self.snapshot_name, self.zfs_node.snapshot_time_format)
        if sys.version_info[0] >= 3:
            from datetime import timezone
            if self.zfs_node.utc:
                dt = dt.replace(tzinfo=timezone.utc)
            seconds = dt.timestamp()
        else:
            # python2 has no good functions to deal with UTC. Yet the unix timestamp
            # must be in UTC to allow comparison against `time.time()` in on other parts
            # of this project (e.g. Thinner.py). If we are handling UTC timestamps,
            # we must adjust for that here.
            if self.zfs_node.utc:
                seconds = (dt - datetime(1970, 1, 1)).total_seconds()
            else:
                seconds = time.mktime(dt.timetuple())
        return seconds

    def from_names(self, names, force_exists=None):
        """convert a list[names] to a list ZfsDatasets for this zfs_node

        Args:
            :type names: list[str]
        """
        ret = []
        for name in names:
            ret.append(self.zfs_node.get_dataset(name, force_exists))

        return ret

    # def add_virtual_snapshot(self, snapshot):
    #     """pretend a snapshot exists (usefull in test mode)"""
    #
    #     # NOTE: we could just call self.snapshots.append() but this would trigger a zfs list which is not always needed.
    #     if CachedProperty.is_cached(self, 'snapshots'):
    #         # already cached so add it
    #         print ("ADDED")
    #         self.snapshots.append(snapshot)
    #     else:
    #         # self.snapshots will add it when requested
    #         print ("ADDED VIRT")
    #         self._virtual_snapshots.append(snapshot)

    @CachedProperty
    def snapshots(self):
        """get all snapshots of this dataset
        :rtype: ZfsDataset
        """

        #FIXME: dont check for existance. (currenlty needed for _add_virtual_snapshots)
        if not self.exists:
            return []

        self.debug("Getting snapshots")

        cmd = [
            "zfs", "list", "-d", "1", "-r", "-t", "snapshot", "-H", "-o", "name", self.name
        ]

        return self.from_names(self.zfs_node.run(cmd=cmd, readonly=True), force_exists=True)

    @property
    def our_snapshots(self):
        """get list[snapshots] creates by us of this dataset"""
        ret = []
        for snapshot in self.snapshots:
            if snapshot.is_ours():
                ret.append(snapshot)

        return ret

    def find_snapshot(self, snapshot):
        """find snapshot by snapshot (can be a snapshot_name or a different
        ZfsDataset )

        Args:
            :rtype: ZfsDataset
            :type snapshot: str or ZfsDataset
        """

        if not isinstance(snapshot, ZfsDataset):
            snapshot_name = snapshot
        else:
            snapshot_name = snapshot.snapshot_name

        for snapshot in self.snapshots:
            if snapshot.snapshot_name == snapshot_name:
                return snapshot

        return None

    def find_snapshot_index(self, snapshot):
        """find snapshot index by snapshot (can be a snapshot_name or
        ZfsDataset)

        Args:
            :type snapshot: str or ZfsDataset
        """

        if not isinstance(snapshot, ZfsDataset):
            snapshot_name = snapshot
        else:
            snapshot_name = snapshot.snapshot_name

        index = 0
        for snapshot in self.snapshots:
            if snapshot.snapshot_name == snapshot_name:
                return index
            index = index + 1

        return None

    @CachedProperty
    def written_since_ours(self):
        """get number of bytes written since our last snapshot"""

        latest_snapshot = self.our_snapshots[-1]

        self.debug("Getting bytes written since our last snapshot")
        cmd = ["zfs", "get", "-H", "-ovalue", "-p", "written@" + str(latest_snapshot), self.name]

        output = self.zfs_node.run(readonly=True, tab_split=False, cmd=cmd, valid_exitcodes=[0])

        return int(output[0])

    def is_changed_ours(self, min_changed_bytes=1):
        """dataset is changed since OUR latest snapshot?

        Args:
            :type min_changed_bytes: int
        """

        if min_changed_bytes == 0:
            return True

        if not self.our_snapshots:
            return True

        # NOTE: filesystems can have a very small amount written without actual changes in some cases
        if self.written_since_ours < min_changed_bytes:
            return False

        return True

    @CachedProperty
    def recursive_datasets(self, types="filesystem,volume"):
        """get all (non-snapshot) datasets recursively under us

        Args:
            :type types: str
        """

        self.debug("Getting all recursive datasets under us")

        names = self.zfs_node.run(tab_split=False, readonly=True, valid_exitcodes=[0], cmd=[
            "zfs", "list", "-r", "-t", types, "-o", "name", "-H", self.name
        ])

        return self.from_names(names[1:], force_exists=True)

    @CachedProperty
    def datasets(self, types="filesystem,volume"):
        """get all (non-snapshot) datasets directly under us

        Args:
            :type types: str
        """

        self.debug("Getting all datasets under us")

        names = self.zfs_node.run(tab_split=False, readonly=True, valid_exitcodes=[0], cmd=[
            "zfs", "list", "-r", "-t", types, "-o", "name", "-H", "-d", "1", self.name
        ])

        return self.from_names(names[1:], force_exists=True)

    def send_pipe(self, features, prev_snapshot, resume_token, show_progress, raw, send_properties, write_embedded,
                  send_pipes, zfs_compressed):
        """returns a pipe with zfs send output for this snapshot

        resume_token: resume sending from this token. (in that case we don't
        need to know snapshot names)

        Args:
            :param send_pipes: output cmd array that will be added to actual zfs send command. (e.g. mbuffer or compression program)
            :type send_pipes: list[str]
            :type features: list[str]
            :type prev_snapshot: ZfsDataset
            :type resume_token: str
            :type show_progress: bool
            :type raw: bool
        """
        # build source command
        cmd = []

        cmd.extend(["zfs", "send", ])

        # all kind of performance options:
        if 'large_blocks' in features and "-L" in self.zfs_node.supported_send_options:
            # large block support (only if recordsize>128k which is seldomly used)
            cmd.append("-L")  # --large-block

        if write_embedded and 'embedded_data' in features and "-e" in self.zfs_node.supported_send_options:
            cmd.append("-e")  # --embed; WRITE_EMBEDDED, more compact stream

        if zfs_compressed and "-c" in self.zfs_node.supported_send_options:
            cmd.append("-c")  # --compressed; use compressed WRITE records

        # raw? (send over encrypted data in its original encrypted form without decrypting)
        if raw:
            cmd.append("--raw")

        # progress output
        if show_progress:
            cmd.append("-v")  # --verbose
            cmd.append("-P")  # --parsable

        # resume a previous send? (don't need more parameters in that case)
        if resume_token:
            cmd.extend(["-t", resume_token])

        else:
            # send properties
            if send_properties:
                cmd.append("-p")  # --props

            # incremental?
            if prev_snapshot:
                cmd.extend(["-i", "@" + prev_snapshot.snapshot_name])

            cmd.append(self.name)

        cmd.extend(send_pipes)

        output_pipe = self.zfs_node.run(cmd, pipe=True, readonly=True)

        return output_pipe

    def recv_pipe(self, pipe, features, recv_pipes, filter_properties=None, set_properties=None, ignore_exit_code=False,
                  force=False):
        """starts a zfs recv for this snapshot and uses pipe as input

        note: you can it both on a snapshot or filesystem object. The
        resulting zfs command is the same, only our object cache is invalidated
        differently.

        Args:
            :param recv_pipes: input cmd array that will be prepended to actual zfs recv command. (e.g. mbuffer or decompression program)
            :type pipe: subprocess.pOpen
            :type features: list[str]
            :type filter_properties: list[str]
            :type set_properties: list[str]
            :type ignore_exit_code: bool
        """

        if set_properties is None:
            set_properties = []

        if filter_properties is None:
            filter_properties = []

        # build target command
        cmd = []

        cmd.extend(recv_pipes)

        cmd.extend(["zfs", "recv"])

        # don't let zfs recv mount everything thats received (even with canmount=noauto!)
        cmd.append("-u")

        for property_ in filter_properties:
            cmd.extend(["-x", property_])

        for property_ in set_properties:
            cmd.extend(["-o", property_])

        # verbose output
        cmd.append("-v")

        if force:
            cmd.append("-F")

        if 'extensible_dataset' in features and "-s" in self.zfs_node.supported_recv_options:
            # support resuming
            self.debug("Enabled resume support")
            cmd.append("-s")

        cmd.append(self.filesystem_name)

        if ignore_exit_code:
            valid_exitcodes = []
        else:
            valid_exitcodes = [0]

        # self.zfs_node.reset_progress()
        self.zfs_node.run(cmd, inp=pipe, valid_exitcodes=valid_exitcodes)

        # invalidate cache
        self.invalidate()

        # in test mode we assume everything was ok and it exists
        if self.zfs_node.readonly:
            self.force_exists = True

        # check if transfer was really ok (exit codes have been wrong before due to bugs in zfs-utils and some
        # errors should be ignored, thats where the ignore_exitcodes is for.)
        if not self.exists:
            self.error("error during transfer")
            raise (Exception("Target doesn't exist after transfer, something went wrong."))

        # at this point we're sure the actual dataset exists
        self.parent.force_exists = True

    def automount(self):
        """Mount the dataset as if one did a zfs mount -a, but only for this dataset
        Failure to mount doesnt result in an exception, but outputs errors to STDERR.

        """

        self.debug("Auto mounting")

        if self.properties['type'] != "filesystem":
            return

        if self.properties['canmount'] != 'on':
            return

        if self.properties['mountpoint'] == 'legacy':
            return

        if self.properties['mountpoint'] == 'none':
            return

        if self.properties['encryption'] != 'off' and self.properties['keystatus'] == 'unavailable':
            return

        self.zfs_node.run(["zfs", "mount", self.name], valid_exitcodes=[0,1])

    def transfer_snapshot(self, target_snapshot, features, prev_snapshot, show_progress,
                          filter_properties, set_properties, ignore_recv_exit_code, resume_token,
                          raw, send_properties, write_embedded, send_pipes, recv_pipes, zfs_compressed, force):
        """transfer this snapshot to target_snapshot. specify prev_snapshot for
        incremental transfer

        connects a send_pipe() to recv_pipe()

        Args:
            :type send_pipes: list[str]
            :type recv_pipes: list[str]
            :type target_snapshot: ZfsDataset
            :type features: list[str]
            :type prev_snapshot: ZfsDataset
            :type show_progress: bool
            :type filter_properties: list[str]
            :type set_properties: list[str]
            :type ignore_recv_exit_code: bool
            :type resume_token: str
            :type raw: bool
        """

        if set_properties is None:
            set_properties = []
        if filter_properties is None:
            filter_properties = []

        self.debug("Transfer snapshot to {}".format(target_snapshot.filesystem_name))

        if resume_token:
            self.verbose("resuming")

        # initial or increment
        if not prev_snapshot:
            self.verbose("-> {} (new)".format(target_snapshot.filesystem_name))
        else:
            # incremental
            self.verbose("-> {}".format(target_snapshot.filesystem_name))

        # do it
        pipe = self.send_pipe(features=features, show_progress=show_progress, prev_snapshot=prev_snapshot,
                              resume_token=resume_token, raw=raw, send_properties=send_properties,
                              write_embedded=write_embedded, send_pipes=send_pipes, zfs_compressed=zfs_compressed)
        target_snapshot.recv_pipe(pipe, features=features, filter_properties=filter_properties,
                                  set_properties=set_properties, ignore_exit_code=ignore_recv_exit_code,
                                  recv_pipes=recv_pipes, force=force)

        # try to automount it, if its the initial transfer
        if not prev_snapshot:
            # in test mode it doesnt actually exist, so dont try to mount it/read properties
            if not target_snapshot.zfs_node.readonly:
                target_snapshot.parent.automount()

    def abort_resume(self):
        """abort current resume state"""
        self.debug("Aborting resume")
        self.zfs_node.run(["zfs", "recv", "-A", self.name])

    def rollback(self):
        """rollback to latest existing snapshot on this dataset"""

        for snapshot in reversed(self.snapshots):
            if snapshot.exists:
                self.debug("Rolling back")
                self.zfs_node.run(["zfs", "rollback", snapshot.name])
                return

    def get_resume_snapshot(self, resume_token):
        """returns snapshot that will be resumed by this resume token (run this
        on source with target-token)

        Args:
            :type resume_token: str
        """
        # use zfs send -n option to determine this
        # NOTE: on smartos stderr, on linux stdout
        (stdout, stderr) = self.zfs_node.run(["zfs", "send", "-t", resume_token, "-n", "-v"], valid_exitcodes=[0, 255],
                                             readonly=True, return_stderr=True)
        if stdout:
            lines = stdout
        else:
            lines = stderr
        for line in lines:
            matches = re.findall("toname = .*@(.*)", line)
            if matches:
                snapshot_name = matches[0]
                snapshot = self.zfs_node.get_dataset(self.filesystem_name + "@" + snapshot_name)
                snapshot.debug("resume token belongs to this snapshot")
                return snapshot

        return None

    def thin_list(self, keeps=None, ignores=None):
        """determines list[snapshots] that should be kept or deleted based on
        the thinning schedule. cull the herd!

        returns: ( keeps, obsoletes )

        Args:
            :param keeps: list[snapshots] to always keep (usually the last)
            :param ignores: snapshots to completely ignore (usually incompatible target snapshots that are going to be destroyed anyway)
            :type keeps: list[ZfsDataset]
            :type ignores: list[ZfsDataset]
        """

        if ignores is None:
            ignores = []
        if keeps is None:
            keeps = []

        snapshots = [snapshot for snapshot in self.our_snapshots if snapshot not in ignores]

        return self.zfs_node.thin(snapshots, keep_objects=keeps)

    def thin(self, skip_holds=False):
        """destroys snapshots according to thin_list, except last snapshot

        Args:
            :type skip_holds: bool
        """

        (keeps, obsoletes) = self.thin_list(keeps=self.our_snapshots[-1:])
        for obsolete in obsoletes:
            if skip_holds and obsolete.is_hold():
                obsolete.verbose("Keeping (common snapshot)")
            else:
                obsolete.destroy()
                self.snapshots.remove(obsolete)

    def find_common_snapshot(self, target_dataset, guid_check):
        """find latest common snapshot between us and target returns None if its
        an initial transfer

        Args:
            :type guid_check: bool
            :type target_dataset: ZfsDataset
        """

        if not target_dataset.snapshots:
            # target has nothing yet
            return None
        else:
            for source_snapshot in reversed(self.snapshots):
                target_snapshot = target_dataset.find_snapshot(source_snapshot)
                if target_snapshot:
                    if guid_check and source_snapshot.properties['guid'] != target_snapshot.properties['guid']:
                        target_snapshot.warning("Common snapshot has invalid guid, ignoring.")
                    else:
                        target_snapshot.debug("common snapshot")
                        return source_snapshot
            # target_dataset.error("Cant find common snapshot with source.")
            raise (Exception("Cant find common snapshot with target."))

    def find_start_snapshot(self, common_snapshot, also_other_snapshots):
        """finds first snapshot to send :rtype: ZfsDataset or None if we cant
        find it.

        Args:
            :type common_snapshot: ZfsDataset
            :type also_other_snapshots: bool
        """

        if not common_snapshot:
            if not self.snapshots:
                start_snapshot = None
            else:
                # no common snapshot, start from beginning
                start_snapshot = self.snapshots[0]

                if not start_snapshot.is_ours() and not also_other_snapshots:
                    # try to start at a snapshot thats ours
                    start_snapshot = self.find_next_snapshot(start_snapshot, also_other_snapshots)
        else:
            # normal situation: start_snapshot is the one after the common snapshot
            start_snapshot = self.find_next_snapshot(common_snapshot, also_other_snapshots)

        return start_snapshot

    def find_incompatible_snapshots(self, common_snapshot, raw):
        """returns a list[snapshots] that is incompatible for a zfs recv onto
        the common_snapshot. all direct followup snapshots with written=0 are
        compatible.

        in raw-mode nothing is compatible. issue #219

        Args:
            :type common_snapshot: ZfsDataset
            :type raw: bool
        """

        ret = []

        if common_snapshot and self.snapshots:
            followup = True
            for snapshot in self.snapshots[self.find_snapshot_index(common_snapshot) + 1:]:
                if raw or not followup or int(snapshot.properties['written']) != 0:
                    followup = False
                    ret.append(snapshot)

        return ret

    def get_allowed_properties(self, filter_properties, set_properties):
        """only returns lists of allowed properties for this dataset type

        Args:
            :type filter_properties: list[str]
            :type set_properties: list[str]
        """

        allowed_filter_properties = []
        allowed_set_properties = []
        illegal_properties = self.ILLEGAL_PROPERTIES[self.properties['type']]
        for set_property in set_properties:
            (property_, value) = set_property.split("=")
            if property_ not in illegal_properties:
                allowed_set_properties.append(set_property)

        for filter_property in filter_properties:
            if filter_property not in illegal_properties:
                allowed_filter_properties.append(filter_property)

        return allowed_filter_properties, allowed_set_properties

    def _add_virtual_snapshots(self, source_dataset, source_start_snapshot, also_other_snapshots):
        """add snapshots from source to our snapshot list. (just the in memory
        list, no disk operations)

        Args:
            :type source_dataset: ZfsDataset
            :type source_start_snapshot: ZfsDataset
            :type also_other_snapshots: bool
        """

        self.debug("Creating virtual target snapshots")
        snapshot = source_start_snapshot
        while snapshot:
            # create virtual target snapsho
            # NOTE: with force_exist we're telling the dataset it doesnt exist yet. (e.g. its virtual)
            virtual_snapshot = self.zfs_node.get_dataset(self.filesystem_name + "@" + snapshot.snapshot_name,
                                                         force_exists=False)
            self.snapshots.append(virtual_snapshot)
            snapshot = source_dataset.find_next_snapshot(snapshot, also_other_snapshots)

    def _pre_clean(self, common_snapshot, target_dataset, source_obsoletes, target_obsoletes, target_keeps):
        """cleanup old stuff before starting snapshot syncing

        Args:
            :type common_snapshot: ZfsDataset
            :type target_dataset: ZfsDataset
            :type source_obsoletes: list[ZfsDataset]
            :type target_obsoletes: list[ZfsDataset]
            :type target_keeps: list[ZfsDataset]
        """

        # on source: destroy all obsoletes before common. (since we cant send them anyways)
        # But after common, only delete snapshots that target also doesn't want
        if common_snapshot:
            before_common = True
        else:
            before_common = False

        for source_snapshot in self.snapshots:
            if common_snapshot and source_snapshot.snapshot_name == common_snapshot.snapshot_name:
                before_common = False
                # never destroy common snapshot
            else:
                target_snapshot = target_dataset.find_snapshot(source_snapshot)
                if (source_snapshot in source_obsoletes) and (before_common or (target_snapshot not in target_keeps)):
                    source_snapshot.destroy()

        # on target: destroy everything thats obsolete, except common_snapshot
        for target_snapshot in target_dataset.snapshots:
            if (target_snapshot in target_obsoletes) \
                    and (not common_snapshot or (target_snapshot.snapshot_name != common_snapshot.snapshot_name)):
                if target_snapshot.exists:
                    target_snapshot.destroy()

    def _validate_resume_token(self, target_dataset, start_snapshot):
        """validate and get (or destory) resume token

        Args:
            :type target_dataset: ZfsDataset
            :type start_snapshot: ZfsDataset
        """

        if target_dataset.exists and 'receive_resume_token' in target_dataset.properties:
            if start_snapshot == None:
                target_dataset.verbose("Aborting resume, its obsolete.")
                target_dataset.abort_resume()
            else:
                resume_token = target_dataset.properties['receive_resume_token']
                # not valid anymore
                resume_snapshot = self.get_resume_snapshot(resume_token)
                if not resume_snapshot or start_snapshot.snapshot_name != resume_snapshot.snapshot_name:
                    target_dataset.verbose("Aborting resume, its no longer valid.")
                    target_dataset.abort_resume()
                else:
                    return resume_token

    def _plan_sync(self, target_dataset, also_other_snapshots, guid_check, raw):
        """plan where to start syncing and what to sync and what to keep

        Args:
            :rtype: ( ZfsDataset, ZfsDataset, list[ZfsDataset], list[ZfsDataset], list[ZfsDataset], list[ZfsDataset] )
            :type target_dataset: ZfsDataset
            :type also_other_snapshots: bool
            :type guid_check: bool
            :type raw: bool
        """

        # determine common and start snapshot
        target_dataset.debug("Determining start snapshot")
        common_snapshot = self.find_common_snapshot(target_dataset, guid_check=guid_check)
        start_snapshot = self.find_start_snapshot(common_snapshot, also_other_snapshots)
        incompatible_target_snapshots = target_dataset.find_incompatible_snapshots(common_snapshot, raw)

        # let thinner decide whats obsolete on source
        source_obsoletes = []
        if self.our_snapshots:
            source_obsoletes = self.thin_list(keeps=[self.our_snapshots[-1]])[1]

        # let thinner decide keeps/obsoletes on target, AFTER the transfer would be done (by using virtual snapshots)
        target_dataset._add_virtual_snapshots(self, start_snapshot, also_other_snapshots)
        target_keeps = []
        target_obsoletes = []
        if target_dataset.our_snapshots:
            (target_keeps, target_obsoletes) = target_dataset.thin_list(keeps=[target_dataset.our_snapshots[-1]],
                                                                        ignores=incompatible_target_snapshots)

        return common_snapshot, start_snapshot, source_obsoletes, target_obsoletes, target_keeps, incompatible_target_snapshots

    def handle_incompatible_snapshots(self, incompatible_target_snapshots, destroy_incompatible):
        """destroy incompatbile snapshots on target before sync, or inform user
        what to do

        Args:
            :type incompatible_target_snapshots: list[ZfsDataset]
            :type destroy_incompatible: bool
        """

        if incompatible_target_snapshots:
            if not destroy_incompatible:
                for snapshot in incompatible_target_snapshots:
                    snapshot.error("Incompatible snapshot")
                raise (Exception("Please destroy incompatible snapshots on target, or use --destroy-incompatible."))
            else:
                for snapshot in incompatible_target_snapshots:
                    snapshot.verbose("Incompatible snapshot")
                    snapshot.destroy(fail_exception=True)
                    self.snapshots.remove(snapshot)

                if len(incompatible_target_snapshots) > 0:
                    self.rollback()

    def sync_snapshots(self, target_dataset, features, show_progress, filter_properties, set_properties,
                       ignore_recv_exit_code, holds, rollback, decrypt, encrypt, also_other_snapshots,
                       no_send, destroy_incompatible, send_pipes, recv_pipes, zfs_compressed, force, guid_check):
        """sync this dataset's snapshots to target_dataset, while also thinning
        out old snapshots along the way.

        Args:
            :type send_pipes: list[str]
            :type recv_pipes: list[str]
            :type target_dataset: ZfsDataset
            :type features: list[str]
            :type show_progress: bool
            :type filter_properties: list[str]
            :type set_properties: list[str]
            :type ignore_recv_exit_code: bool
            :type holds: bool
            :type rollback: bool
            :type decrypt: bool
            :type also_other_snapshots: bool
            :type no_send: bool
            :type guid_check: bool
        """

        # self.verbose("-> {}".format(target_dataset))

        # defaults for these settings if there is no encryption stuff going on:
        send_properties = True
        raw = False
        write_embedded = True

        # source dataset encrypted?
        if self.properties.get('encryption', 'off') != 'off':
            # user wants to send it over decrypted?
            if decrypt:
                # when decrypting, zfs cant send properties
                send_properties = False
            else:
                # keep data encrypted by sending it raw (including properties)
                raw = True

        (common_snapshot, start_snapshot, source_obsoletes, target_obsoletes, target_keeps,
         incompatible_target_snapshots) = \
            self._plan_sync(target_dataset=target_dataset, also_other_snapshots=also_other_snapshots,
                            guid_check=guid_check, raw=raw)

        # NOTE: we do this because we dont want filesystems to fillup when backups keep failing.
        # Also usefull with no_send to still cleanup stuff.
        self._pre_clean(
            common_snapshot=common_snapshot, target_dataset=target_dataset,
            target_keeps=target_keeps, target_obsoletes=target_obsoletes, source_obsoletes=source_obsoletes)

        # handle incompatible stuff on target
        target_dataset.handle_incompatible_snapshots(incompatible_target_snapshots, destroy_incompatible)

        # now actually transfer the snapshots, if we want
        if no_send:
            return

        # check if we can resume
        resume_token = self._validate_resume_token(target_dataset, start_snapshot)

        (active_filter_properties, active_set_properties) = self.get_allowed_properties(filter_properties,
                                                                                        set_properties)

        # encrypt at target?
        if encrypt and not raw:
            # filter out encryption properties to let encryption on the target take place
            active_filter_properties.extend(["keylocation", "pbkdf2iters", "keyformat", "encryption"])
            write_embedded = False

        # now actually transfer the snapshots
        prev_source_snapshot = common_snapshot
        source_snapshot = start_snapshot
        do_rollback = rollback
        while source_snapshot:
            target_snapshot = target_dataset.find_snapshot(source_snapshot)  # still virtual

            # does target actually want it?
            if target_snapshot not in target_obsoletes and not source_snapshot.is_excluded:

                # do the rollback, one time at first transfer
                if do_rollback:
                    target_dataset.rollback()
                    do_rollback = False

                source_snapshot.transfer_snapshot(target_snapshot, features=features,
                                                  prev_snapshot=prev_source_snapshot, show_progress=show_progress,
                                                  filter_properties=active_filter_properties,
                                                  set_properties=active_set_properties,
                                                  ignore_recv_exit_code=ignore_recv_exit_code,
                                                  resume_token=resume_token, write_embedded=write_embedded, raw=raw,
                                                  send_properties=send_properties, send_pipes=send_pipes,
                                                  recv_pipes=recv_pipes, zfs_compressed=zfs_compressed, force=force)

                resume_token = None

                # hold the new common snapshots and release the previous ones
                if holds:
                    target_snapshot.hold()
                    source_snapshot.hold()

                if prev_source_snapshot:
                    if holds:
                        prev_source_snapshot.release()
                        target_dataset.find_snapshot(prev_source_snapshot).release()

                # we may now destroy the previous source snapshot if its obsolete
                if prev_source_snapshot in source_obsoletes:
                    prev_source_snapshot.destroy()

                # destroy the previous target snapshot if obsolete (usually this is only the common_snapshot,
                # the rest was already destroyed or will not be send)
                prev_target_snapshot = target_dataset.find_snapshot(prev_source_snapshot)
                if prev_target_snapshot in target_obsoletes:
                    prev_target_snapshot.destroy()

                prev_source_snapshot = source_snapshot
            else:
                source_snapshot.debug("skipped (target doesn't need it)")
                # was it actually a resume?
                if resume_token:
                    target_dataset.verbose("Aborting resume, we dont want that snapshot anymore.")
                    target_dataset.abort_resume()
                    resume_token = None

            source_snapshot = self.find_next_snapshot(source_snapshot, also_other_snapshots)

    def mount(self, mount_point):

        self.debug("Mounting")

        cmd = [
            "mount", "-tzfs", self.name, mount_point
        ]

        self.zfs_node.run(cmd=cmd, valid_exitcodes=[0])

    def unmount(self, mount_point):

        self.debug("Unmounting")

        cmd = [
            "umount", mount_point
        ]

        self.zfs_node.run(cmd=cmd, valid_exitcodes=[0])

    def clone(self, name):
        """clones this snapshot and returns ZfsDataset of the clone"""

        self.debug("Cloning to {}".format(name))

        cmd = [
            "zfs", "clone", self.name, name
        ]

        self.zfs_node.run(cmd=cmd, valid_exitcodes=[0])

        return self.zfs_node.get_dataset(name, force_exists=True)

    def set(self, prop, value):
        """set a zfs property"""

        self.debug("Setting {}={}".format(prop, value))

        cmd = [
            "zfs", "set", "{}={}".format(prop, value), self.name
        ]

        self.zfs_node.run(cmd=cmd, valid_exitcodes=[0])

        self.invalidate()

    def inherit(self, prop):
        """inherit zfs property"""

        self.debug("Inheriting property {}".format(prop))

        cmd = [
            "zfs", "inherit", prop, self.name
        ]

        self.zfs_node.run(cmd=cmd, valid_exitcodes=[0])

        self.invalidate()



from basetest import *
import time

class TestZfsAutobackup31(unittest2.TestCase):
    """various new 3.1 features"""

    def setUp(self):
        prepare_zpools()
        self.longMessage=True

    def test_no_thinning(self):

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty".split(" ")).run())

        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty --keep-target=0 --keep-source=0 --no-thinning".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1@test-20101111000001
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source1/fs1/sub@test-20101111000001
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs2/sub@test-20101111000001
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source1/fs1/sub@test-20101111000001
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
test_target1/test_source2/fs2/sub@test-20101111000001
""")


    def test_re_replication(self):
        """test re-replication of something thats already a backup (new in v3.1-beta5)"""

        shelltest("zfs create test_target1/a")
        shelltest("zfs create test_target1/b")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1/a --no-progress --verbose --debug".split(" ")).run())

        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup("test test_target1/b --no-progress --verbose".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t snapshot test_target1")
            #NOTE: it wont backup test_target1/a/test_source2/fs2/sub to test_target1/b since it doesnt have the zfs_autobackup property anymore.
            self.assertMultiLineEqual(r,"""
test_target1/a/test_source1/fs1@test-20101111000000
test_target1/a/test_source1/fs1/sub@test-20101111000000
test_target1/a/test_source2/fs2/sub@test-20101111000000
test_target1/b/test_source1/fs1@test-20101111000000
test_target1/b/test_source1/fs1/sub@test-20101111000000
test_target1/b/test_source2/fs2/sub@test-20101111000000
test_target1/b/test_target1/a/test_source1/fs1@test-20101111000000
test_target1/b/test_target1/a/test_source1/fs1/sub@test-20101111000000
""")

    def test_zfs_compressed(self):

        with mocktime("20101111000000"):
            self.assertFalse(
                ZfsAutobackup("test test_target1 --no-progress --verbose --debug --zfs-compressed".split(" ")).run())

    def test_force(self):
        """test 1:1 replication"""

        shelltest("zfs set autobackup:test=true test_source1")

        with mocktime("20101111000000"):
            self.assertFalse(
                ZfsAutobackup("test test_target1 --no-progress --verbose --debug --force --strip-path=1".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t snapshot test_target1")
            self.assertMultiLineEqual(r,"""
test_target1@test-20101111000000
test_target1/fs1@test-20101111000000
test_target1/fs1/sub@test-20101111000000
test_target1/fs2/sub@test-20101111000000
""")


    def test_exclude_unchanged(self):

        shelltest("zfs snapshot -r test_source1@somesnapshot")

        with mocktime("20101111000000"):
            self.assertFalse(
                ZfsAutobackup(
                    "test test_target1 --verbose --allow-empty --exclude-unchanged=1".split(" ")).run())

        #everything should be excluded, but should not return an error (see #190)
        with mocktime("20101111000001"):
            self.assertFalse(
                ZfsAutobackup(
                    "test test_target1 --verbose --allow-empty --exclude-unchanged=1".split(" ")).run())

        r = shelltest("zfs list -H -o name -r -t snapshot test_target1")
        self.assertMultiLineEqual(r, """
test_target1/test_source2/fs2/sub@test-20101111000000
""")




from zfs_autobackup.CmdPipe import CmdPipe
from basetest import *
import time

# We have to do a LOT to properly test encryption/decryption/raw transfers
#
# For every scenario we need at least:
# - plain source dataset
# - encrypted source dataset
# - plain target path
# - encrypted target path
# - do a full transfer
# - do a incremental transfer

# Scenarios:
# - Raw transfer
# - Decryption transfer (--decrypt)
# - Encryption transfer (--encrypt)
# - Re-encryption transfer (--decrypt --encrypt)

class TestZfsEncryption(unittest2.TestCase):


    def setUp(self):
        prepare_zpools()

        try:
            shelltest("zfs get encryption test_source1")
        except:
            self.skipTest("Encryption not supported on this ZFS version.")

    def load_key(self, key, path):

        shelltest("rm /tmp/zfstest.key 2>/dev/null;true")
        shelltest("echo {} > /tmp/zfstest.key".format(key))
        shelltest("zfs load-key {}".format(path))

    def prepare_encrypted_dataset(self, key, path, unload_key=False):

        # create encrypted source dataset
        shelltest("rm /tmp/zfstest.key 2>/dev/null;true")
        shelltest("echo {} > /tmp/zfstest.key".format(key))
        shelltest("zfs create -o keylocation=file:///tmp/zfstest.key -o keyformat=passphrase -o encryption=on {}".format(path))

        if unload_key:
            shelltest("zfs unmount {}".format(path))
            shelltest("zfs unload-key {}".format(path))

        # r=shelltest("dd if=/dev/zero of=/test_source1/fs1/enc1/data.txt bs=200000 count=1")

    def  test_raw(self):
        """send encrypted data unaltered (standard operation)"""

        self.prepare_encrypted_dataset("11111111", "test_source1/fs1/encryptedsource")
        self.prepare_encrypted_dataset("11111111", "test_source1/fs1/encryptedsourcekeyless", unload_key=True) # raw mode shouldn't need a key
        self.prepare_encrypted_dataset("22222222", "test_target1/encryptedtarget")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-progress --allow-empty --exclude-received".split(" ")).run())
            self.assertFalse(ZfsAutobackup("test test_target1/encryptedtarget --verbose --no-progress --no-snapshot --exclude-received".split(" ")).run())

        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-progress --allow-empty --exclude-received".split(" ")).run())
            self.assertFalse(ZfsAutobackup("test test_target1/encryptedtarget --verbose --no-progress --no-snapshot --exclude-received".split(" ")).run())

        r = shelltest("zfs get -r -t filesystem encryptionroot test_target1")
        self.assertMultiLineEqual(r,"""
NAME                                                                  PROPERTY        VALUE                                                                 SOURCE
test_target1                                                          encryptionroot  -                                                                     -
test_target1/encryptedtarget                                          encryptionroot  test_target1/encryptedtarget                                          -
test_target1/encryptedtarget/test_source1                             encryptionroot  test_target1/encryptedtarget                                          -
test_target1/encryptedtarget/test_source1/fs1                         encryptionroot  -                                                                     -
test_target1/encryptedtarget/test_source1/fs1/encryptedsource         encryptionroot  test_target1/encryptedtarget/test_source1/fs1/encryptedsource         -
test_target1/encryptedtarget/test_source1/fs1/encryptedsourcekeyless  encryptionroot  test_target1/encryptedtarget/test_source1/fs1/encryptedsourcekeyless  -
test_target1/encryptedtarget/test_source1/fs1/sub                     encryptionroot  -                                                                     -
test_target1/encryptedtarget/test_source2                             encryptionroot  test_target1/encryptedtarget                                          -
test_target1/encryptedtarget/test_source2/fs2                         encryptionroot  test_target1/encryptedtarget                                          -
test_target1/encryptedtarget/test_source2/fs2/sub                     encryptionroot  -                                                                     -
test_target1/test_source1                                             encryptionroot  -                                                                     -
test_target1/test_source1/fs1                                         encryptionroot  -                                                                     -
test_target1/test_source1/fs1/encryptedsource                         encryptionroot  test_target1/test_source1/fs1/encryptedsource                         -
test_target1/test_source1/fs1/encryptedsourcekeyless                  encryptionroot  test_target1/test_source1/fs1/encryptedsourcekeyless                  -
test_target1/test_source1/fs1/sub                                     encryptionroot  -                                                                     -
test_target1/test_source2                                             encryptionroot  -                                                                     -
test_target1/test_source2/fs2                                         encryptionroot  -                                                                     -
test_target1/test_source2/fs2/sub                                     encryptionroot  -                                                                     -
""")

    def  test_decrypt(self):
        """decrypt data and store unencrypted (--decrypt)"""

        self.prepare_encrypted_dataset("11111111", "test_source1/fs1/encryptedsource")
        self.prepare_encrypted_dataset("22222222", "test_target1/encryptedtarget")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-progress --decrypt --allow-empty --exclude-received".split(" ")).run())
            self.assertFalse(ZfsAutobackup("test test_target1/encryptedtarget --verbose --no-progress --decrypt --no-snapshot --exclude-received".split(" ")).run())

        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-progress --decrypt --allow-empty --exclude-received".split(" ")).run())
            self.assertFalse(ZfsAutobackup("test test_target1/encryptedtarget --verbose --no-progress --decrypt --no-snapshot --exclude-received".split(" ")).run())

        r = shelltest("zfs get -r -t filesystem encryptionroot test_target1")
        self.assertEqual(r, """
NAME                                                           PROPERTY        VALUE                         SOURCE
test_target1                                                   encryptionroot  -                             -
test_target1/encryptedtarget                                   encryptionroot  test_target1/encryptedtarget  -
test_target1/encryptedtarget/test_source1                      encryptionroot  test_target1/encryptedtarget  -
test_target1/encryptedtarget/test_source1/fs1                  encryptionroot  -                             -
test_target1/encryptedtarget/test_source1/fs1/encryptedsource  encryptionroot  -                             -
test_target1/encryptedtarget/test_source1/fs1/sub              encryptionroot  -                             -
test_target1/encryptedtarget/test_source2                      encryptionroot  test_target1/encryptedtarget  -
test_target1/encryptedtarget/test_source2/fs2                  encryptionroot  test_target1/encryptedtarget  -
test_target1/encryptedtarget/test_source2/fs2/sub              encryptionroot  -                             -
test_target1/test_source1                                      encryptionroot  -                             -
test_target1/test_source1/fs1                                  encryptionroot  -                             -
test_target1/test_source1/fs1/encryptedsource                  encryptionroot  -                             -
test_target1/test_source1/fs1/sub                              encryptionroot  -                             -
test_target1/test_source2                                      encryptionroot  -                             -
test_target1/test_source2/fs2                                  encryptionroot  -                             -
test_target1/test_source2/fs2/sub                              encryptionroot  -                             -
""")

    def  test_encrypt(self):
        """send normal data set and store encrypted on the other side (--encrypt) issue #60 """

        self.prepare_encrypted_dataset("11111111", "test_source1/fs1/encryptedsource")
        self.prepare_encrypted_dataset("22222222", "test_target1/encryptedtarget")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-progress --encrypt --debug --allow-empty --exclude-received --clear-mountpoint".split(" ")).run())
            self.assertFalse(ZfsAutobackup("test test_target1/encryptedtarget --verbose --no-progress --encrypt --debug --no-snapshot --exclude-received --clear-mountpoint".split(" ")).run())

        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-progress --encrypt --debug --allow-empty --exclude-received --clear-mountpoint".split(" ")).run())
            self.assertFalse(ZfsAutobackup("test test_target1/encryptedtarget --verbose --no-progress --encrypt --debug --no-snapshot --exclude-received --clear-mountpoint".split(" ")).run())

        r = shelltest("zfs get -r -t filesystem encryptionroot test_target1")
        self.assertEqual(r, """
NAME                                                           PROPERTY        VALUE                                                          SOURCE
test_target1                                                   encryptionroot  -                                                              -
test_target1/encryptedtarget                                   encryptionroot  test_target1/encryptedtarget                                   -
test_target1/encryptedtarget/test_source1                      encryptionroot  test_target1/encryptedtarget                                   -
test_target1/encryptedtarget/test_source1/fs1                  encryptionroot  test_target1/encryptedtarget                                   -
test_target1/encryptedtarget/test_source1/fs1/encryptedsource  encryptionroot  test_target1/encryptedtarget/test_source1/fs1/encryptedsource  -
test_target1/encryptedtarget/test_source1/fs1/sub              encryptionroot  test_target1/encryptedtarget                                   -
test_target1/encryptedtarget/test_source2                      encryptionroot  test_target1/encryptedtarget                                   -
test_target1/encryptedtarget/test_source2/fs2                  encryptionroot  test_target1/encryptedtarget                                   -
test_target1/encryptedtarget/test_source2/fs2/sub              encryptionroot  test_target1/encryptedtarget                                   -
test_target1/test_source1                                      encryptionroot  -                                                              -
test_target1/test_source1/fs1                                  encryptionroot  -                                                              -
test_target1/test_source1/fs1/encryptedsource                  encryptionroot  test_target1/test_source1/fs1/encryptedsource                  -
test_target1/test_source1/fs1/sub                              encryptionroot  -                                                              -
test_target1/test_source2                                      encryptionroot  -                                                              -
test_target1/test_source2/fs2                                  encryptionroot  -                                                              -
test_target1/test_source2/fs2/sub                              encryptionroot  -                                                              -
""")

    def test_reencrypt(self):
        """reencrypt data (--decrypt --encrypt) """

        self.prepare_encrypted_dataset("11111111", "test_source1/fs1/encryptedsource")
        self.prepare_encrypted_dataset("22222222", "test_target1/encryptedtarget")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup(
                "test test_target1 --verbose --no-progress --decrypt --encrypt --debug --allow-empty --exclude-received --clear-mountpoint".split(" ")).run())
            self.assertFalse(ZfsAutobackup(
                "test test_target1/encryptedtarget --verbose --no-progress --decrypt --encrypt --debug --no-snapshot --exclude-received --clear-mountpoint".split(
                    " ")).run())

        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup(
                "test test_target1 --verbose --no-progress --decrypt --encrypt --debug --allow-empty --exclude-received".split(" ")).run())
            self.assertFalse(ZfsAutobackup(
                "test test_target1/encryptedtarget --verbose --no-progress --decrypt --encrypt --debug --no-snapshot --exclude-received".split(
                    " ")).run())

        r = shelltest("zfs get -r -t filesystem encryptionroot test_target1")
        self.assertEqual(r, """
NAME                                                           PROPERTY        VALUE                         SOURCE
test_target1                                                   encryptionroot  -                             -
test_target1/encryptedtarget                                   encryptionroot  test_target1/encryptedtarget  -
test_target1/encryptedtarget/test_source1                      encryptionroot  test_target1/encryptedtarget  -
test_target1/encryptedtarget/test_source1/fs1                  encryptionroot  test_target1/encryptedtarget  -
test_target1/encryptedtarget/test_source1/fs1/encryptedsource  encryptionroot  test_target1/encryptedtarget  -
test_target1/encryptedtarget/test_source1/fs1/sub              encryptionroot  test_target1/encryptedtarget  -
test_target1/encryptedtarget/test_source2                      encryptionroot  test_target1/encryptedtarget  -
test_target1/encryptedtarget/test_source2/fs2                  encryptionroot  test_target1/encryptedtarget  -
test_target1/encryptedtarget/test_source2/fs2/sub              encryptionroot  test_target1/encryptedtarget  -
test_target1/test_source1                                      encryptionroot  -                             -
test_target1/test_source1/fs1                                  encryptionroot  -                             -
test_target1/test_source1/fs1/encryptedsource                  encryptionroot  -                             -
test_target1/test_source1/fs1/sub                              encryptionroot  -                             -
test_target1/test_source2                                      encryptionroot  -                             -
test_target1/test_source2/fs2                                  encryptionroot  -                             -
test_target1/test_source2/fs2/sub                              encryptionroot  -                             -
""")




    def  test_raw_invalid_snapshot(self):
        """in raw mode, its not allowed to have any newer snaphots on target, #219"""

        self.prepare_encrypted_dataset("11111111", "test_source1/fs1/encryptedsource")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-progress".split(" ")).run())

        #this is invalid in raw mode
        shelltest("zfs snapshot test_target1/test_source1/fs1/encryptedsource@incompatible")

        with mocktime("20101111000001"):
            #should fail because of incompatble snapshot
            self.assertEqual(ZfsAutobackup("test test_target1 --verbose --no-progress --allow-empty".split(" ")).run(),1)
            #should destroy incompatible and continue
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-progress --no-snapshot --destroy-incompatible".split(" ")).run())


        r = shelltest("zfs get -r -t filesystem encryptionroot test_target1")
        self.assertMultiLineEqual(r,"""
NAME                                           PROPERTY        VALUE                                          SOURCE
test_target1                                   encryptionroot  -                                              -
test_target1/test_source1                      encryptionroot  -                                              -
test_target1/test_source1/fs1                  encryptionroot  -                                              -
test_target1/test_source1/fs1/encryptedsource  encryptionroot  test_target1/test_source1/fs1/encryptedsource  -
test_target1/test_source1/fs1/sub              encryptionroot  -                                              -
test_target1/test_source2                      encryptionroot  -                                              -
test_target1/test_source2/fs2                  encryptionroot  -                                              -
test_target1/test_source2/fs2/sub              encryptionroot  -                                              -
""")


    def  test_resume_encrypt_with_no_key(self):
        """test what happens if target encryption key not loaded (this led to a kernel crash of freebsd with 2.1.x i think) while trying to resume"""

        self.prepare_encrypted_dataset("11111111", "test_source1/fs1/encryptedsource")
        self.prepare_encrypted_dataset("22222222", "test_target1/encryptedtarget")


        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1/encryptedtarget --verbose --no-progress --encrypt --allow-empty --exclude-received --clear-mountpoint".split(" ")).run())

        r = shelltest("zfs set compress=off test_source1 test_target1")

        # big change on source
        r = shelltest("dd if=/dev/zero of=/test_source1/fs1/data bs=250M count=1")

        # waste space on target
        r = shelltest("dd if=/dev/zero of=/test_target1/waste bs=250M count=1")

        # should fail and leave resume token
        with mocktime("20101111000001"):
            self.assertTrue(ZfsAutobackup(
                "test test_target1/encryptedtarget --verbose --no-progress --encrypt --exclude-received --allow-empty --clear-mountpoint".split(
                    " ")).run())
        #
        # free up space
        r = shelltest("rm /test_target1/waste")

        # sync
        r = shelltest("zfs umount test_target1")
        r = shelltest("zfs mount test_target1")

        #
        # #unload key
        shelltest("zfs unload-key test_target1/encryptedtarget")

        # resume should fail
        with mocktime("20101111000001"):
            self.assertEqual(ZfsAutobackup(
                "test test_target1/encryptedtarget --verbose --no-progress --encrypt --exclude-received --allow-empty --no-snapshot --clear-mountpoint".split(
                    " ")).run(),3)



#NOTE: On some versions this leaves 2 weird sub-datasets that should'nt be there (its probably a zfs bug?)
#so we ignore this, and just make sure the backup resumes correctly after reloading the key.
#         r = shelltest("zfs get -r -t all encryptionroot test_target1")
#         self.assertEqual(r, """
# NAME                                                                               PROPERTY        VALUE                                                          SOURCE
# test_target1                                                                       encryptionroot  -                                                              -
# test_target1/encryptedtarget                                                       encryptionroot  test_target1/encryptedtarget                                   -
# test_target1/encryptedtarget/test_source1                                          encryptionroot  test_target1/encryptedtarget                                   -
# test_target1/encryptedtarget/test_source1/fs1                                      encryptionroot  test_target1/encryptedtarget                                   -
# test_target1/encryptedtarget/test_source1/fs1@test-20101111000000                  encryptionroot  test_target1/encryptedtarget                                   -
# test_target1/encryptedtarget/test_source1/fs1/encryptedsource                      encryptionroot  test_target1/encryptedtarget/test_source1/fs1/encryptedsource  -
# test_target1/encryptedtarget/test_source1/fs1/encryptedsource@test-20101111000000  encryptionroot  test_target1/encryptedtarget/test_source1/fs1/encryptedsource  -
# test_target1/encryptedtarget/test_source1/fs1/encryptedsource@test-20101111000001  encryptionroot  test_target1/encryptedtarget/test_source1/fs1/encryptedsource  -
# test_target1/encryptedtarget/test_source1/fs1/sub                                  encryptionroot  test_target1/encryptedtarget                                   -
# test_target1/encryptedtarget/test_source1/fs1/sub@test-20101111000000              encryptionroot  test_target1/encryptedtarget                                   -
# test_target1/encryptedtarget/test_source1/fs1/sub/sub                              encryptionroot  -                                                              -
# test_target1/encryptedtarget/test_source1/fs1/sub/sub@test-20101111000001          encryptionroot  -                                                              -
# test_target1/encryptedtarget/test_source2                                          encryptionroot  test_target1/encryptedtarget                                   -
# test_target1/encryptedtarget/test_source2/fs2                                      encryptionroot  test_target1/encryptedtarget                                   -
# test_target1/encryptedtarget/test_source2/fs2/sub                                  encryptionroot  test_target1/encryptedtarget                                   -
# test_target1/encryptedtarget/test_source2/fs2/sub@test-20101111000000              encryptionroot  test_target1/encryptedtarget                                   -
# test_target1/encryptedtarget/test_source2/fs2/sub/sub                              encryptionroot  -                                                              -
# test_target1/encryptedtarget/test_source2/fs2/sub/sub@test-20101111000001          encryptionroot  -                                                              -
# """)



        #reload key and resume correctly.
        self.load_key("22222222", "test_target1/encryptedtarget")

        # resume should complete
        with mocktime("20101111000001"):
            self.assertEqual(ZfsAutobackup(
                "test test_target1/encryptedtarget --verbose --no-progress --encrypt --exclude-received --allow-empty --no-snapshot --clear-mountpoint".split(
                    " ")).run(),0)




from basetest import *

from zfs_autobackup.ExecuteNode import ExecuteNode

run_orig=ExecuteNode.run
run_counter=0

def run_count(*args, **kwargs):
    global run_counter
    run_counter=run_counter+1
    return (run_orig(*args, **kwargs))

class TestZfsScaling(unittest2.TestCase):

    def setUp(self):
        prepare_zpools()
        self.longMessage = True

    def test_manysnapshots(self):
        """count the number of commands when there are many snapshots."""

        snapshot_count=100

        print("Creating many snapshots...")
        s=""
        for i in range(1970,1970+snapshot_count):
            s=s+"zfs snapshot test_source1/fs1@test-{:04}1111000000;".format(i)

        shelltest(s)

        global run_counter

        run_counter=0
        with patch.object(ExecuteNode,'run', run_count) as p:

            with mocktime("20101112000000"):
                self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --keep-source=10000 --keep-target=10000 --no-holds --allow-empty".split(" ")).run())


            #this triggers if you make a change with an impact of more than O(snapshot_count/2)
            expected_runs=342
            print("EXPECTED RUNS: {}".format(expected_runs))
            print("ACTUAL RUNS  : {}".format(run_counter))
            self.assertLess(abs(run_counter-expected_runs), snapshot_count/2)


        run_counter=0
        with patch.object(ExecuteNode,'run', run_count) as p:

            with mocktime("20101112000001"):
                self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --keep-source=10000 --keep-target=10000 --no-holds --allow-empty".split(" ")).run())


            #this triggers if you make a change with a performance impact of more than O(snapshot_count/2)
            expected_runs=47
            print("EXPECTED RUNS: {}".format(expected_runs))
            print("ACTUAL RUNS  : {}".format(run_counter))
            self.assertLess(abs(run_counter-expected_runs), snapshot_count/2)

    def test_manydatasets(self):
        """count the number of commands when when there are many datasets"""

        dataset_count=100

        print("Creating many datasets...")
        s=""
        for i in range(0,dataset_count):
            s=s+"zfs create test_source1/fs1/{};".format(i)

        shelltest(s)

        global run_counter

        #first run
        run_counter=0
        with patch.object(ExecuteNode,'run', run_count) as p:

            with mocktime("20101112000000"):
                self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-holds --allow-empty".split(" ")).run())


            #this triggers if you make a change with an impact of more than O(snapshot_count/2)`
            expected_runs=842
            print("EXPECTED RUNS: {}".format(expected_runs))
            print("ACTUAL RUNS: {}".format(run_counter))
            self.assertLess(abs(run_counter-expected_runs), dataset_count/2)


        #second run, should have higher number of expected_runs
        run_counter=0
        with patch.object(ExecuteNode,'run', run_count) as p:

            with mocktime("20101112000001"):
                self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-holds --allow-empty".split(" ")).run())


            #this triggers if you make a change with a performance impact of more than O(snapshot_count/2)
            expected_runs=1047
            print("EXPECTED RUNS: {}".format(expected_runs))
            print("ACTUAL RUNS: {}".format(run_counter))
            self.assertLess(abs(run_counter-expected_runs), dataset_count/2)
